# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
cimport async_channel.channels as channels
cimport async_channel.consumer as consumers
cimport async_channel.producer as producers

cdef class ExchangeChannel(channels.Channel):
    cdef public object exchange_manager
    cdef public object exchange

    cdef int filter_send_counter
    cdef bint should_send_filter

    cpdef object get_filtered_consumers(self, str cryptocurrency=*, str symbol=*)

cdef class TimeFrameExchangeChannel(ExchangeChannel):
    cpdef object get_filtered_consumers(self, str cryptocurrency=*, str symbol=*, str time_frame=*)

cdef class ExchangeChannelConsumer(consumers.Consumer):
    pass

cdef class ExchangeChannelProducer(producers.Producer):
    cpdef void trigger_single_update(self)

cdef class ExchangeChannelInternalConsumer(consumers.InternalConsumer):
    pass

cdef class ExchangeChannelSupervisedConsumer(consumers.SupervisedConsumer):
    pass

cpdef ExchangeChannel get_chan(str chan_name, str exchange_id)
cpdef dict get_exchange_channels(str exchange_id)
cpdef void set_chan(ExchangeChannel chan, str name)
cpdef void del_exchange_channel_container(str exchange_id)
cpdef void del_chan(str name, str exchange_id)



# pylint: disable=E0611
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import time

import octobot_trading.errors as errors

import octobot_trading.constants as constants
import octobot_trading.exchange_data.kline.channel.kline as kline_channel
import octobot_trading.enums as enums


class KlineUpdater(kline_channel.KlineProducer):
    CHANNEL_NAME = constants.KLINE_CHANNEL
    KLINE_REFRESH_TIME = 8
    QUICK_KLINE_REFRESH_TIME = 3

    def __init__(self, channel):
        super().__init__(channel)
        self.refresh_time = KlineUpdater.KLINE_REFRESH_TIME
        self.tasks = []

    async def start(self):
        """
        Creates OHLCV refresh tasks
        """
        refresh_threshold = self.channel.exchange_manager.get_rest_pairs_refresh_threshold()
        if refresh_threshold is enums.RestExchangePairsRefreshMaxThresholds.MEDIUM:
            self.refresh_time = 14
        elif refresh_threshold is enums.RestExchangePairsRefreshMaxThresholds.SLOW:
            self.refresh_time = 22
        if self.channel.is_paused:
            await self.pause()
        else:
            self.tasks = [
                asyncio.create_task(self.time_frame_watcher(time_frame))
                for time_frame in self.channel.exchange_manager.exchange_config.traded_time_frames]

    async def time_frame_watcher(self, time_frame):
        """
        Manage timeframe OHLCV data refreshing for all pairs
        """
        while not self.should_stop and not self.channel.is_paused:
            try:
                started_time = time.time()
                quick_sleep = False
                for pair in self.channel.exchange_manager.exchange_config.traded_symbol_pairs:
                    candle: list = await self.channel.exchange_manager.exchange.get_kline_price(pair, time_frame)
                    try:
                        candle = candle[0]
                        await self.push(time_frame, pair, candle)
                    except TypeError:
                        pass
                    except IndexError:
                        quick_sleep = True
                        self.logger.debug(f"Not enough data to compute kline data in {time_frame} for {pair}. "
                                          f"Kline will be updated with the next refresh.")

                sleep_time = max((self.QUICK_KLINE_REFRESH_TIME if quick_sleep else self.refresh_time)
                                 - (time.time() - started_time), 0)
                await asyncio.sleep(sleep_time)
            except errors.FailedRequest as e:
                self.logger.warning(str(e))
                # avoid spamming on disconnected situation
                await asyncio.sleep(constants.DEFAULT_FAILED_REQUEST_RETRY_TIME)
            except errors.NotSupported:
                self.logger.warning(f"{self.channel.exchange_manager.exchange_name} is not supporting updates")
                await self.pause()
            except Exception as e:
                self.logger.exception(e, True, f"Failed to update kline data in {time_frame} : {e}")

    async def resume(self) -> None:
        await super().resume()
        if not self.is_running:
            for task in self.tasks:
                task.cancel()
            await self.run()



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

cdef class Trade:
    cdef public object trader
    cdef public object exchange_manager

    cdef public object side # TradeOrderSide
    cdef public object status # OrderStatus
    cdef public object trade_type # TraderOrderType

    cdef public str symbol
    cdef public str currency
    cdef public str market
    cdef public str taker_or_maker
    cdef public str trade_id
    cdef public str origin_order_id
    cdef public bint simulated
    cdef public bint is_closing_order
    cdef public bint reduce_only
    cdef public str tag
    cdef public list associated_entry_ids

    cdef public object origin_price
    cdef public object origin_quantity
    cdef public object executed_quantity
    cdef public object executed_price
    cdef public object total_cost
    cdef public object trade_profitability
    cdef public str quantity_currency

    cdef public double timestamp
    cdef public double creation_time
    cdef public double canceled_time
    cdef public double executed_time

    cdef public dict fee # Dict[str, Union[str, decimal.Decimal]]

    cdef public object exchange_trade_type # raw exchange trade type, used to create trade dict

    cpdef double get_time(self)
    cpdef object get_quantity(self)
    cpdef bint has_been_executed(self)
    cpdef void update_from_order(self,
                                 object order,
                                 double canceled_time=*,
                                 double creation_time=*,
                                 double executed_time=*)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_trading.enums as trading_enums
import octobot_trading.exchanges as exchanges
import octobot_trading.constants as trading_constants


def create_order_signal_content(
        order, action, strategy, exchange_manager,
        target_amount=None,
        target_position=None,
        updated_target_amount=None,
        updated_target_position=None,
        updated_limit_price=trading_constants.ZERO,
        updated_stop_price=trading_constants.ZERO,
        updated_current_price=trading_constants.ZERO,
) -> dict:
    return {
        trading_enums.TradingSignalCommonsAttrs.ACTION.value: action.value,
        trading_enums.TradingSignalOrdersAttrs.STRATEGY.value: strategy,
        trading_enums.TradingSignalOrdersAttrs.SYMBOL.value: order.symbol,
        trading_enums.TradingSignalOrdersAttrs.EXCHANGE.value: exchange_manager.exchange_name,
        trading_enums.TradingSignalOrdersAttrs.EXCHANGE_TYPE.value: exchanges.get_exchange_type(exchange_manager).value,
        trading_enums.TradingSignalOrdersAttrs.SIDE.value: order.side.value,
        trading_enums.TradingSignalOrdersAttrs.TYPE.value: order.order_type.value,
        trading_enums.TradingSignalOrdersAttrs.QUANTITY.value: float(order.origin_quantity),
        trading_enums.TradingSignalOrdersAttrs.TARGET_AMOUNT.value: target_amount
        if updated_target_amount is None else updated_target_amount,
        trading_enums.TradingSignalOrdersAttrs.TARGET_POSITION.value: target_position
        if updated_target_position is None else updated_target_position,
        trading_enums.TradingSignalOrdersAttrs.UPDATED_TARGET_AMOUNT.value: updated_target_amount,
        trading_enums.TradingSignalOrdersAttrs.UPDATED_TARGET_POSITION.value: updated_target_position,
        trading_enums.TradingSignalOrdersAttrs.LIMIT_PRICE.value: float(updated_limit_price)
        if updated_limit_price else float(order.origin_price),
        trading_enums.TradingSignalOrdersAttrs.UPDATED_LIMIT_PRICE.value: float(updated_limit_price),
        trading_enums.TradingSignalOrdersAttrs.STOP_PRICE.value: float(updated_stop_price)
        if updated_stop_price else float(order.origin_stop_price),
        trading_enums.TradingSignalOrdersAttrs.UPDATED_STOP_PRICE.value: float(updated_stop_price),
        trading_enums.TradingSignalOrdersAttrs.CURRENT_PRICE.value: float(updated_current_price)
        if updated_current_price else float(order.created_last_price),
        trading_enums.TradingSignalOrdersAttrs.UPDATED_CURRENT_PRICE.value: float(updated_current_price),
        trading_enums.TradingSignalOrdersAttrs.REDUCE_ONLY.value: order.reduce_only,
        trading_enums.TradingSignalOrdersAttrs.POST_ONLY.value: False,
        trading_enums.TradingSignalOrdersAttrs.GROUP_ID.value:
            None if order.order_group is None else order.order_group.name,
        trading_enums.TradingSignalOrdersAttrs.GROUP_TYPE.value:
            None if order.order_group is None else order.order_group.__class__.__name__,
        trading_enums.TradingSignalOrdersAttrs.TAG.value: order.tag,
        trading_enums.TradingSignalOrdersAttrs.ASSOCIATED_ORDER_IDS.value: order.associated_entry_ids,
        trading_enums.TradingSignalOrdersAttrs.SHARED_SIGNAL_ORDER_ID.value: order.shared_signal_order_id,
        trading_enums.TradingSignalOrdersAttrs.BUNDLED_WITH.value:
            None if order.triggered_by is None else order.triggered_by.shared_signal_order_id
        if order.has_been_bundled else None,
        trading_enums.TradingSignalOrdersAttrs.CHAINED_TO.value:
            None if order.triggered_by is None else order.triggered_by.shared_signal_order_id,
        trading_enums.TradingSignalOrdersAttrs.ADDITIONAL_ORDERS.value: [],
    }



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import mock

import octobot_trading.exchanges.types as exchange_types
import pytest

from tests.exchanges import exchange_manager


@pytest.fixture
def websocket_exchange(exchange_manager):
    return exchange_types.WebSocketExchange(exchange_manager.config, exchange_manager)


def test_clear(websocket_exchange):
    assert websocket_exchange.exchange_manager is not None
    assert websocket_exchange.exchange is not None
    websocket_exchange.websocket_connectors = [mock.Mock(), mock.Mock()]

    websocket_exchange.clear()
    assert websocket_exchange.exchange_manager is None
    assert websocket_exchange.exchange is None
    websocket_exchange.websocket_connectors[0].clear.assert_called_once()
    websocket_exchange.websocket_connectors[1].clear.assert_called_once()



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_trading.personal_data.positions cimport position_state
from octobot_trading.personal_data.positions.position_state cimport (
    PositionState,
)

from octobot_trading.personal_data.positions cimport position
from octobot_trading.personal_data.positions.position cimport (
    Position,
)

from octobot_trading.personal_data.positions cimport types
from octobot_trading.personal_data.positions.types cimport (
    LinearPosition,
    InversePosition,
)

from octobot_trading.personal_data.positions cimport states
from octobot_trading.personal_data.positions.states cimport (
    LiquidatePositionState,
    IdlePositionState,
    ActivePositionState,
)

from octobot_trading.personal_data.positions cimport channel
from octobot_trading.personal_data.positions.channel cimport (
    PositionsProducer,
    PositionsChannel,
    PositionsUpdater,
    PositionsUpdaterSimulator,
)

from octobot_trading.personal_data.positions cimport positions_manager
from octobot_trading.personal_data.positions.positions_manager cimport (
    PositionsManager,
)

from octobot_trading.personal_data.positions cimport position_util
from octobot_trading.personal_data.positions.position_util cimport (
    parse_position_status,
)

from octobot_trading.personal_data.positions cimport position_factory
from octobot_trading.personal_data.positions.position_factory cimport (
    create_position_instance_from_raw,
    create_position_from_type,
    create_symbol_position,
)

__all__ = [
    "PositionState",
    "PositionsProducer",
    "PositionsChannel",
    "PositionsUpdaterSimulator",
    "Position",
    "LinearPosition",
    "InversePosition",
    "PositionsUpdater",
    "PositionsManager",
    "create_position_instance_from_raw",
    "create_position_from_type",
    "create_symbol_position",
    "parse_position_status",
    "LiquidatePositionState",
    "IdlePositionState",
    "ActivePositionState",
]



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_trading.exchange_data.prices import channel
from octobot_trading.exchange_data.prices import prices_manager
from octobot_trading.exchange_data.prices import price_events_manager

from octobot_trading.exchange_data.prices.channel import (
    MarkPriceUpdater,
    MarkPriceUpdaterSimulator,
    MarkPriceProducer,
    MarkPriceChannel,
)
from octobot_trading.exchange_data.prices.prices_manager import (
    PricesManager,
    calculate_mark_price_from_recent_trade_prices,
)
from octobot_trading.exchange_data.prices.price_events_manager import (
    PriceEventsManager,
)

__all__ = [
    "MarkPriceUpdaterSimulator",
    "MarkPriceProducer",
    "MarkPriceChannel",
    "PricesManager",
    "calculate_mark_price_from_recent_trade_prices",
    "MarkPriceUpdater",
    "PriceEventsManager",
]



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_trading.enums as enums
import octobot_trading.personal_data.orders.types.limit.limit_order as limit_order


class StopLossOrder(limit_order.LimitOrder):
    def __init__(self, trader, side=enums.TradeOrderSide.SELL):
        super().__init__(trader, side)
        self.trigger_above = self.side is enums.TradeOrderSide.BUY

    def is_counted_in_available_funds(self):
        return False

    def _filled_maker_or_taker(self):
        # Creates a market order when filled, which is taker
        return enums.ExchangeConstantsMarketPropertyColumns.TAKER.value

    async def on_filled(self):
        await limit_order.LimitOrder.on_filled(self)
        if not self.trader.simulate and self.is_self_managed():
            # TODO replace with chained order ?
            await self.trader.create_artificial_order(enums.TraderOrderType.SELL_MARKET
                                                      if self.side is enums.TradeOrderSide.SELL
                                                      else enums.TraderOrderType.BUY_MARKET,
                                                      self.symbol, self.origin_stop_price,
                                                      self.origin_quantity, self.origin_stop_price)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import copy
import decimal

import octobot_commons.logging as logging
import octobot_commons.symbols as symbol_util

import octobot_trading.constants as constants
import octobot_trading.errors as errors


class PortfolioValueHolder:
    """
    PortfolioValueHolder calculates the current and the origin portfolio value in reference market for each updates
    """

    def __init__(self, portfolio_manager):
        self.portfolio_manager = portfolio_manager
        self.logger = logging.get_logger(f"{self.__class__.__name__}"
                                         f"[{self.portfolio_manager.exchange_manager.exchange_name}]")

        self.initializing_symbol_prices = set()
        self.initializing_symbol_prices_pairs = set()

        self.portfolio_origin_value = constants.ZERO
        self.portfolio_current_value = constants.ZERO

        # values in decimal.Decimal
        self.last_prices_by_trading_pair = {}
        self.origin_portfolio = None

        # values in decimal.Decimal
        self.origin_crypto_currencies_values = {}
        self.current_crypto_currencies_values = {}

        # set of currencies for which the current exchange is not providing any suitable price data
        self.missing_currency_data_in_exchange = set()

    def reset_portfolio_values(self):
        self.portfolio_origin_value = constants.ZERO
        self.portfolio_current_value = constants.ZERO

        self.origin_portfolio = None

        self.origin_crypto_currencies_values = {}
        self.current_crypto_currencies_values = {}

    def update_origin_crypto_currencies_values(self, symbol, mark_price):
        """
        Update origin cryptocurrencies value
        :param symbol: the symbol to update
        :param mark_price: the symbol mark price value in decimal.Decimal
        :return: True if the origin portfolio should be recomputed
        """
        currency, market = symbol_util.parse_symbol(symbol).base_and_quote()
        # update origin values if this price has relevant data regarding the origin portfolio (using both quote and base)
        origin_currencies_should_be_updated = (
                (
                        currency not in set(self.origin_crypto_currencies_values.keys()) and
                        market == self.portfolio_manager.reference_market
                )
                or
                (
                        market not in set(self.origin_crypto_currencies_values.keys()) and
                        currency == self.portfolio_manager.reference_market
                )
        )
        if origin_currencies_should_be_updated:
            # will fail if symbol doesn't have a price in self.origin_crypto_currencies_values and therefore
            # requires the origin portfolio value to be recomputed using this price info in case this price is relevant
            if market == self.portfolio_manager.reference_market:
                self.origin_crypto_currencies_values[currency] = mark_price
            else:
                self.origin_crypto_currencies_values[market] = constants.ONE / mark_price
        self.last_prices_by_trading_pair[symbol] = mark_price
        return origin_currencies_should_be_updated

    def get_current_crypto_currencies_values(self):
        """
        Return the current crypto-currencies values
        :return: the current crypto-currencies values
        """
        if not self.current_crypto_currencies_values:
            self._update_portfolio_and_currencies_current_value()
        return self.current_crypto_currencies_values

    def get_current_holdings_values(self):
        """
        Get holdings ratio for each currencies
        :return: the holdings ratio dictionary
        """
        holdings = self.get_current_crypto_currencies_values()
        return {currency: self._get_currency_value(self.portfolio_manager.portfolio.portfolio, currency, holdings)
                for currency in holdings.keys()}

    def get_currency_holding_ratio(self, currency):
        """
        Return the holdings ratio for the specified currency
        :param currency: the currency
        :return: the holdings ratio
        """
        return self._evaluate_value(currency,
                                    self.portfolio_manager.portfolio.get_currency_portfolio(
                                        currency).total) / self.portfolio_current_value \
            if self.portfolio_current_value else constants.ZERO

    def handle_profitability_recalculation(self, force_recompute_origin_portfolio):
        """
        Initialize values required by portfolio profitability to perform its profitability calculation
        :param force_recompute_origin_portfolio: when True, force origin portfolio computation
        """
        self._update_portfolio_and_currencies_current_value()
        self._init_portfolio_values_if_necessary(force_recompute_origin_portfolio)

    def get_origin_portfolio_current_value(self, refresh_values=False):
        """
        Calculates and return the origin portfolio actual value
        :param refresh_values: when True, force origin portfolio reevaluation
        :return: the origin portfolio current value
        """
        if refresh_values:
            self.current_crypto_currencies_values.update(
                self._evaluate_config_crypto_currencies_and_portfolio_values(self.origin_portfolio.portfolio))
        return self._update_portfolio_current_value(self.origin_portfolio.portfolio,
                                                    currencies_values=self.current_crypto_currencies_values)

    def _init_portfolio_values_if_necessary(self, force_recompute_origin_portfolio):
        """
        Init origin portfolio values if necessary
        :param force_recompute_origin_portfolio: when True, force origin portfolio computation
        """
        if self.portfolio_origin_value == constants.ZERO:
            # try to update portfolio origin value if it's not known yet
            self._init_origin_portfolio_and_currencies_value()
        if force_recompute_origin_portfolio:
            self._recompute_origin_portfolio_initial_value()

    def _init_origin_portfolio_and_currencies_value(self):
        """
        Initialize origin portfolio and the origin portfolio currencies values
        """
        self.origin_portfolio = self.origin_portfolio or copy.copy(self.portfolio_manager.portfolio)
        self.origin_crypto_currencies_values.update(
            self._evaluate_config_crypto_currencies_and_portfolio_values(self.origin_portfolio.portfolio,
                                                                         ignore_missing_currency_data=True))
        self._recompute_origin_portfolio_initial_value()

    def _update_portfolio_current_value(self, portfolio, currencies_values=None, fill_currencies_values=False):
        """
        Update the portfolio with current prices
        :param portfolio: the portfolio to update
        :param currencies_values: the currencies values
        :param fill_currencies_values: the currencies values to calculate
        :return: the updated portfolio
        """
        values = currencies_values
        if values is None or fill_currencies_values:
            self.current_crypto_currencies_values.update(
                self._evaluate_config_crypto_currencies_and_portfolio_values(portfolio))
            if fill_currencies_values:
                self._fill_currencies_values(currencies_values)
            values = self.current_crypto_currencies_values
        return self._evaluate_portfolio_value(portfolio, values)

    def _fill_currencies_values(self, currencies_values):
        """
        Fill a currency values dict with new data
        :param currencies_values: currencies values dict to be filled
        """
        currencies_values.update({
            currency: value
            for currency, value in self.current_crypto_currencies_values.items()
            if currency not in currencies_values
        })

    def _update_portfolio_and_currencies_current_value(self):
        """
        Update the portfolio current value with the current portfolio instance
        """
        self.portfolio_current_value = self._update_portfolio_current_value(
            self.portfolio_manager.portfolio.portfolio)

    def _evaluate_value(self, currency, quantity, raise_error=True):
        """
        Evaluate value returns the currency quantity value in the reference (attribute) currency
        :param currency: the currency to evaluate
        :param quantity: the currency quantity
        :param raise_error: will catch exception if False
        :return: the currency value
        """
        # easy case --> the current currency is the reference currency or the quantity is 0
        if currency == self.portfolio_manager.reference_market or quantity == constants.ZERO:
            return quantity
        currency_value = self._try_get_value_of_currency(currency, quantity, raise_error)
        return self._check_currency_initialization(currency=currency, currency_value=currency_value)

    def _check_currency_initialization(self, currency, currency_value):
        """
        Check if the currency has to be removed from self.initializing_symbol_prices and return currency_value
        :param currency: the currency to check
        :param currency_value: the currency value
        :return: currency_value after checking
        """
        if currency_value > constants.ZERO and currency in self.initializing_symbol_prices:
            self.initializing_symbol_prices.remove(currency)
        return currency_value

    def _recompute_origin_portfolio_initial_value(self):
        """
        Compute origin portfolio initial value and update portfolio_origin_value
        """
        if self.portfolio_manager.historical_portfolio_value_manager is not None \
           and self.portfolio_manager.historical_portfolio_value_manager.has_historical_starting_portfolio_value(
            self.portfolio_manager.reference_market
        ):
            # get origin value from history when possible
            value = self.portfolio_manager.historical_portfolio_value_manager.\
                get_historical_starting_starting_portfolio_value(self.portfolio_manager.reference_market)
            if value is not None:
                self.portfolio_origin_value = value
                return
        self.portfolio_origin_value = \
            self._update_portfolio_current_value(self.origin_portfolio.portfolio,
                                                 currencies_values=self.origin_crypto_currencies_values,
                                                 fill_currencies_values=True)

    def _try_get_value_of_currency(self, currency, quantity, raise_error):
        """
        try_get_value_of_currency will try to obtain the current value of the currency quantity in reference currency.
        It will try to create the symbol that fit with the exchange logic.
        :return: the value found of this currency quantity, if not found returns 0.
        """
        try:
            return self.convert_currency_value_using_last_prices(quantity, currency,
                                                                 self.portfolio_manager.reference_market)
        except errors.MissingPriceDataError as missing_data_exception:
            symbol = symbol_util.merge_currencies(currency, self.portfolio_manager.reference_market)
            reversed_symbol = symbol_util.merge_currencies(self.portfolio_manager.reference_market, currency)
            if not any(self.portfolio_manager.exchange_manager.symbol_exists(s) for s in (symbol, reversed_symbol)) \
               and currency not in self.missing_currency_data_in_exchange:
                self._inform_no_matching_symbol(currency)
                self.missing_currency_data_in_exchange.add(currency)
            if not self.portfolio_manager.exchange_manager.is_backtesting:
                self._try_to_ask_ticker_missing_symbol_data(currency, symbol, reversed_symbol)
                if raise_error:
                    raise missing_data_exception
        return constants.ZERO

    def convert_currency_value_using_last_prices(self, quantity, current_currency, target_currency):
        try:
            symbol = symbol_util.merge_currencies(current_currency, target_currency)
            if self._has_price_data(symbol):
                return quantity * self._get_last_price_data(symbol)
        except KeyError:
            pass
        try:
            reversed_symbol = symbol_util.merge_currencies(target_currency, current_currency)
            return quantity / self._get_last_price_data(reversed_symbol)
        except decimal.DivisionByZero:
            pass
        except KeyError:
            pass
        raise errors.MissingPriceDataError(f"no price data to evaluate {current_currency} price in {target_currency}")

    def _has_price_data(self, symbol):
        return self._get_last_price_data(symbol) is not constants.ZERO

    def _get_last_price_data(self, symbol):
        try:
            return self.last_prices_by_trading_pair[symbol]
        except KeyError:
            # a settlement asset or other symbol extra data might be different, try to ignore it
            to_find_symbol = symbol_util.parse_symbol(symbol)
            for symbol_key in self.last_prices_by_trading_pair:
                if symbol_util.parse_symbol(symbol_key).is_same_base_and_quote(to_find_symbol):
                    return self.last_prices_by_trading_pair[symbol_key]
        raise KeyError(symbol)

    def _try_to_ask_ticker_missing_symbol_data(self, currency, symbol, reversed_symbol):
        """
        Try to ask the ticker producer to watch additional symbols
        to collect missing data required for profitability calculation
        :param currency: the concerned currency
        :param symbol: the symbol to add
        :param reversed_symbol: the reversed symbol to add
        """
        symbols_to_add = []
        if self.portfolio_manager.exchange_manager.symbol_exists(symbol):
            symbols_to_add = [symbol]
        elif self.portfolio_manager.exchange_manager.symbol_exists(reversed_symbol):
            symbols_to_add = [reversed_symbol]

        new_symbols_to_add = [s for s in symbols_to_add if s not in self.initializing_symbol_prices_pairs]
        if new_symbols_to_add:
            self.logger.debug(f"Fetching price for {new_symbols_to_add} to compute all the "
                              f"currencies values and properly evaluate portfolio.")
            self._ask_ticker_data_for_currency(new_symbols_to_add)
            self.initializing_symbol_prices.add(currency)
            self.initializing_symbol_prices_pairs.update(new_symbols_to_add)

    def _ask_ticker_data_for_currency(self, symbols_to_add):
        """
        Synchronously call TICKER_CHANNEL producer to add a list of new symbols to its watch list
        :param symbols_to_add: the list of symbol to add to the TICKER_CHANNEL producer watch list
        """
        asyncio.run_coroutine_threadsafe(
            self.portfolio_manager.exchange_manager.exchange_config.add_watched_symbols(symbols_to_add),
            asyncio.get_running_loop())

    def _inform_no_matching_symbol(self, currency):
        """
        Log a missing currency pair to calculate the portfolio profitability
        :param currency: the concerned currency
        """
        # do not log warning in backtesting or tests
        if not self.portfolio_manager.exchange_manager.is_backtesting:
            self.logger.warning(f"No trading pair including {currency} and {self.portfolio_manager.reference_market} on"
                                f" {self.portfolio_manager.exchange_manager.exchange_name}. {currency} "
                                f"can't be valued for portfolio and profitability.")

    def _evaluate_config_crypto_currencies_and_portfolio_values(self,
                                                                portfolio,
                                                                ignore_missing_currency_data=False):
        """
        Evaluate both config and portfolio currencies values
        :param portfolio: the current portfolio
        :param ignore_missing_currency_data: when True, ignore missing currencies values in calculation
        :return: the result of config and portfolio currencies values calculation
        """
        evaluated_pair_values = {}
        evaluated_currencies = set()
        missing_tickers = set()

        self._evaluate_config_currencies_values(evaluated_pair_values, evaluated_currencies, missing_tickers)
        self._evaluate_portfolio_currencies_values(portfolio, evaluated_pair_values, evaluated_currencies,
                                                   missing_tickers, ignore_missing_currency_data)
        return evaluated_pair_values

    def _evaluate_config_currencies_values(self, evaluated_pair_values, evaluated_currencies, missing_tickers):
        """
        Evaluate config currencies values
        TODO do not use config[CONFIG_CRYPTO_CURRENCIES]
        :param evaluated_pair_values: the list of evaluated pairs
        :param evaluated_currencies: the list of evaluated currencies
        :param missing_tickers: the list of missing currencies
        """
        if self.portfolio_manager.exchange_manager.exchange_config.all_config_symbol_pairs:
            currency, market = symbol_util.parse_symbol(
                self.portfolio_manager.exchange_manager.exchange_config.all_config_symbol_pairs[0]
            ).base_and_quote()
            currency_to_evaluate = currency
            try:
                if currency not in evaluated_currencies:
                    evaluated_pair_values[currency] = self._evaluate_value(currency, constants.ONE)
                    evaluated_currencies.add(currency)
                if market not in evaluated_currencies:
                    currency_to_evaluate = market
                    evaluated_pair_values[market] = self._evaluate_value(market, constants.ONE)
                    evaluated_currencies.add(market)
            except errors.MissingPriceDataError:
                missing_tickers.add(currency_to_evaluate)

    def _evaluate_portfolio_currencies_values(self,
                                              portfolio,
                                              evaluated_pair_values,
                                              evaluated_currencies,
                                              missing_tickers,
                                              ignore_missing_currency_data):
        """
        Evaluate current portfolio currencies values
        :param portfolio: the current portfolio
        :param evaluated_pair_values: the list of evaluated pairs
        :param evaluated_currencies: the list of evaluated currencies
        :param missing_tickers: the list of missing currencies
        :param ignore_missing_currency_data: when True, ignore missing currencies values in calculation
        """
        for currency in portfolio:
            try:
                if currency not in evaluated_currencies and self._should_currency_be_considered(
                        currency, portfolio, ignore_missing_currency_data):
                    evaluated_pair_values[currency] = self._evaluate_value(currency, constants.ONE)
                    evaluated_currencies.add(currency)
            except errors.MissingPriceDataError:
                missing_tickers.add(currency)

    def _evaluate_portfolio_value(self, portfolio, currencies_values=None):
        """
        Perform evaluate_value with a portfolio configuration
        :param portfolio: the portfolio to explore
        :param currencies_values: currencies to evaluate
        :return: the calculated quantity value in reference (attribute) currency
        """
        return sum([
            self._get_currency_value(portfolio, currency, currencies_values)
            for currency in portfolio
            if currency not in self.missing_currency_data_in_exchange
        ])

    def _get_currency_value(self, portfolio, currency, currencies_values=None, raise_error=False):
        """
        Return the currency value
        :param portfolio: the specified portfolio
        :param currency: the currency to evaluate
        :param currencies_values: currencies values dict
        :param raise_error: When True, forward exceptions
        :return: the currency value
        """
        if currency in portfolio and portfolio[currency].total != constants.ZERO:
            if currencies_values and currency in currencies_values:
                return currencies_values[currency] * portfolio[currency].total
            return self._evaluate_value(currency, portfolio[currency].total, raise_error)
        return constants.ZERO

    def _should_currency_be_considered(self, currency, portfolio, ignore_missing_currency_data):
        """
        Return True if enough data is available to evaluate currency value
        :param currency: the currency to evaluate
        :param portfolio: the specified portfolio
        :param ignore_missing_currency_data: When True, ignore check of currency presence
        in missing_currency_data_in_exchange
        :return: True if enough data is available to evaluate currency value
        """
        return (currency not in self.missing_currency_data_in_exchange or ignore_missing_currency_data) and \
               (portfolio[currency].total > constants.ZERO or currency in
                self.portfolio_manager.portfolio_profitability.valuated_currencies)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_trading.modes cimport channel
from octobot_trading.modes.channel cimport (
    ModeChannelConsumer,
    ModeChannelProducer,
    ModeChannel,
    AbstractTradingModeConsumer,
    AbstractTradingModeProducer,
)

from octobot_trading.modes cimport modes_util
from octobot_trading.modes.modes_util cimport (
    get_required_candles_count,
)

from octobot_trading.modes cimport modes_factory
from octobot_trading.modes.modes_factory cimport (
    create_temporary_trading_mode_with_local_config,
)

__all__ = [
    "ModeChannelConsumer",
    "ModeChannelProducer",
    "ModeChannel",
    "AbstractTradingModeProducer",
    "AbstractTradingModeConsumer",
    "get_required_candles_count",
    "create_temporary_trading_mode_with_local_config",
]



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
cimport octobot_trading.exchanges.abstract_websocket_exchange as abstract_websocket_exchange

cdef class CCXTWebsocketConnector(abstract_websocket_exchange.AbstractWebsocketExchange):
    cdef public list filtered_pairs
    cdef public list watched_pairs
    cdef public object min_timeframe
    cdef dict _previous_open_candles
    cdef dict _subsequent_unordered_candles_count
    cdef object _start_time_millis
    cdef str websocket_name

    cdef public object local_loop

    cdef public bint should_stop
    cdef public bint is_authenticated
    cdef public object adapter
    cdef public object additional_config
    cdef public dict headers
    cdef public dict options
    cdef public dict feed_tasks
    cdef public object _reconnect_task
    cdef public double _last_close_time
    cdef public double throttled_ws_updates

    # return object when an exception might be thrown
    cpdef object add_headers(self, dict headers_dict)
    cpdef object add_options(self, dict options_dict)

    cdef object _create_client(self)
    cdef bint _has_authenticated_channel(self)
    cdef bint _is_authenticated_feed(self, object feed)
    cdef object _subscribe_feeds(self)
    cdef bint _should_use_authenticated_feeds(self)
    cdef bint _is_supported_channel(self, object channel)
    cdef object _subscribe_candle_feed(self)
    cdef object _subscribe_channels_feeds(self, bint pairs_related_channels_only)
    cdef object _subscribe_pair_independent_feed(self)
    cdef object _subscribe_traded_pairs_feed(self)
    cdef object _subscribe_watched_pairs_feed(self)
    cdef dict _get_feed_generator_by_feed(self)
    cdef object _get_generator(self, str method_name)
    cdef dict _get_callback_by_feed(self)
    cdef object _get_since_filter_value(self, object feed, str time_frame)
    cdef object _subscribe_feed(self, object feed, list symbols=*, str time_frame=*,
                                object since=*, object limit=*, object params=*)
    cdef str _get_feed_identifier(self, object feed_generator, dict kwargs)
    cdef object _filter_exchange_pairs_and_timeframes(self)
    cdef object _add_pair(self, str pair, bint watching_only)
    cdef object _add_exchange_symbols(self)
    cdef object _filter_exchange_symbols(self)
    cdef object _add_time_frame(self, list filtered_timeframes, object time_frame, bint log_on_error)
    cdef object _init_exchange_time_frames(self)
    cdef bint _should_run_candle_feed(self)
    cdef bint _is_supported_pair(self, str pair)
    cdef bint _is_supported_time_frame(self, object time_frame)
    cdef bint _is_pair_independent_feed(self, object feed)
    cdef list _convert_book_prices_to_orders(self, object book_prices_and_volumes, str book_side)
    cdef void _register_previous_open_candle(self, str time_frame, str symbol, list candle)
    cdef list _get_previous_open_candle(self, str time_frame, str symbol)
    cdef void _register_subsequent_unordered_candle(self, str time_frame, str symbol, object parsed_timeframe, double current_candle_time)
    cdef int _get_subsequent_unordered_candles_count(self, str time_frame, str symbol)



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
cimport octobot_trading.exchange_data.prices.price_events_manager as price_events_manager
cimport octobot_trading.exchange_data.order_book.order_book_manager as order_book_manager
cimport octobot_trading.exchange_data.prices.prices_manager as prices_manager
cimport octobot_trading.exchange_data.recent_trades.recent_trades_manager as recent_trades_manager
cimport octobot_trading.exchange_data.ticker.ticker_manager as ticker_manager
cimport octobot_trading.exchange_data.funding.funding_manager as funding_manager

cdef class ExchangeSymbolData:
    cdef public str symbol

    cdef object logger
    cdef public object exchange_manager

    cdef public dict symbol_candles
    cdef public dict symbol_klines

    cdef public price_events_manager.PriceEventsManager price_events_manager
    cdef public order_book_manager.OrderBookManager order_book_manager
    cdef public prices_manager.PricesManager prices_manager
    cdef public recent_trades_manager.RecentTradesManager recent_trades_manager
    cdef public ticker_manager.TickerManager ticker_manager
    cdef public funding_manager.FundingManager funding_manager

    cpdef list handle_recent_trade_update(self, list recent_trades, bint replace_all=*)
    cpdef void handle_order_book_update(self, list asks, list bids)
    cpdef void handle_order_book_ticker_update(self, double ask_quantity, double ask_price,
                                               double bid_quantity, double bid_price)
    cpdef bint handle_mark_price_update(self, object mark_price, str mark_price_source)
    cpdef void handle_ticker_update(self, dict ticker)
    cpdef void handle_mini_ticker_update(self, dict mini_ticker)
    cpdef void stop(self)



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.


""" Order class will represent an open order in the specified exchange
In simulation it will also define rules to be filled / canceled
It is also use to store creation & fill values of the order """
cimport octobot_trading.personal_data.portfolios.portfolio_manager as portfolio_manager
cimport octobot_trading.personal_data.portfolios.portfolio_value_holder as portfolio_value_holder

cdef class PortfolioProfitability:
    cdef object logger

    cdef portfolio_manager.PortfolioManager portfolio_manager
    cdef portfolio_value_holder.PortfolioValueHolder value_manager

    cdef public object profitability
    cdef public object profitability_percent
    cdef public object profitability_diff
    cdef public object market_profitability_percent
    cdef public object initial_portfolio_current_profitability

    cdef set traded_currencies_without_market_specific
    cdef public set valuated_currencies

    cpdef object update_profitability(self, bint force_recompute_origin_portfolio=*)
    cpdef void reset_profitability(self)

    cdef object _calculate_average_market_profitability(self)
    cdef void _reset_before_profitability_calculation(self)
    cdef object _update_profitability_calculation(self)
    cdef object _update_portfolio_delta(self)
    cdef dict _only_symbol_currency_filter(self, dict currency_dict)
    # cdef void _init_traded_currencies_without_market_specific(self) can't be cythonized for now



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio

import octobot_trading.exchange_channel as exchanges_channel
import octobot_trading.enums as enums


class PositionsProducer(exchanges_channel.ExchangeChannelProducer):
    async def push(self, positions):
        await self.perform(positions)

    async def perform(self, positions):
        try:
            for position in positions:
                if not position:
                    continue
                symbol: str = position[enums.ExchangeConstantsPositionColumns.SYMBOL.value]
                changed = await self.channel.exchange_manager.exchange_personal_data. \
                    handle_position_update(symbol=symbol,
                                           side=position[enums.ExchangeConstantsPositionColumns.SIDE.value],
                                           position=position,
                                           should_notify=False)

                if changed:
                    await self.send(cryptocurrency=self.channel.exchange_manager.exchange.
                                    get_pair_cryptocurrency(symbol),
                                    symbol=symbol,
                                    position=position,
                                    is_updated=changed)
        except asyncio.CancelledError:
            self.logger.info("Update tasks cancelled.")
        except Exception as e:
            self.logger.exception(e, True, f"Exception when triggering update: {e}")

    async def send(self, cryptocurrency, symbol, position, is_updated=False):
        for consumer in self.channel.get_filtered_consumers(symbol=symbol):
            await consumer.queue.put({
                "exchange": self.channel.exchange_manager.exchange_name,
                "exchange_id": self.channel.exchange_manager.id,
                "cryptocurrency": cryptocurrency,
                "symbol": symbol,
                "position": position,
                "is_updated": is_updated
            })


class PositionsChannel(exchanges_channel.ExchangeChannel):
    PRODUCER_CLASS = PositionsProducer
    CONSUMER_CLASS = exchanges_channel.ExchangeChannelConsumer



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import pytest
from octobot_backtesting.backtesting import Backtesting

from octobot_commons.tests.test_config import load_test_config
from octobot_trading.exchanges.exchange_manager import ExchangeManager
from octobot_trading.api.exchange import cancel_ccxt_throttle_task
from octobot_trading.exchanges.types import RestExchange
from octobot_trading.exchanges.implementations import ExchangeSimulator
from tests import event_loop

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


class TestExchangeManager:
    EXCHANGE_NAME = "binanceus"

    @staticmethod
    async def init_default(config=None, simulated=True, backtesting=False):
        if not config:
            config = load_test_config()

        exchange_manager = ExchangeManager(config, TestExchangeManager.EXCHANGE_NAME)
        exchange_manager.is_simulated = simulated
        exchange_manager.is_backtesting = backtesting
        if backtesting:
            exchange_manager.backtesting = Backtesting(None, [exchange_manager.id], None, [])
        exchange_manager.rest_only = True

        await exchange_manager.initialize()
        return config, exchange_manager

    async def test_create_exchange(self):
        # simulated
        config, exchange_manager = await self.init_default(simulated=True, backtesting=True)

        assert exchange_manager is not None
        assert exchange_manager.exchange_name is TestExchangeManager.EXCHANGE_NAME

        assert exchange_manager.exchange.authenticated() is False

        assert exchange_manager.config is config

        assert isinstance(exchange_manager.exchange, ExchangeSimulator)
        await exchange_manager.stop()

        # real
        config, exchange_manager = await self.init_default(simulated=False)

        assert exchange_manager is not None
        assert exchange_manager.exchange_name is TestExchangeManager.EXCHANGE_NAME

        assert exchange_manager.config is config

        assert isinstance(exchange_manager.exchange, RestExchange)
        cancel_ccxt_throttle_task()
        await exchange_manager.stop()

    async def test_get_is_overloaded(self):
        # simulated
        config, exchange_manager = await self.init_default(simulated=True, backtesting=True)

        # few traded elements
        exchange_manager.exchange_config.traded_symbol_pairs = [0] * 3
        exchange_manager.exchange_config.traded_time_frames = [0] * 3
        assert exchange_manager.get_currently_handled_pair_with_time_frame() == 3*3
        assert not exchange_manager.get_is_overloaded()

        # too many traded elements
        exchange_manager.exchange_config.traded_symbol_pairs = [0] * 99
        exchange_manager.exchange_config.traded_time_frames = [0] * 99
        assert exchange_manager.get_currently_handled_pair_with_time_frame() == 99*99
        # not overloaded since exchange simulator is not limited in term of traded elements
        assert not exchange_manager.get_is_overloaded()
        await exchange_manager.stop()

        # real
        config, exchange_manager = await self.init_default(simulated=False)

        # few traded elements
        exchange_manager.exchange_config.traded_symbol_pairs = [0] * 3
        exchange_manager.exchange_config.traded_time_frames = [0] * 3
        assert exchange_manager.get_currently_handled_pair_with_time_frame() == 3*3
        assert not exchange_manager.get_is_overloaded()

        # too many traded elements
        exchange_manager.exchange_config.traded_symbol_pairs = [0] * 99
        exchange_manager.exchange_config.traded_time_frames = [0] * 99
        assert exchange_manager.get_currently_handled_pair_with_time_frame() == 99*99
        # overloaded since exchange simulator is limited in term of traded elements
        assert exchange_manager.get_is_overloaded()
        cancel_ccxt_throttle_task()
        await exchange_manager.stop()

    async def test_ready(self):
        _, exchange_manager = await self.init_default()

        assert exchange_manager.is_ready
        cancel_ccxt_throttle_task()
        await exchange_manager.stop()



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
cimport octobot_trading.exchanges.exchange_manager as exchange_manager

cdef class AbstractWebsocketExchange:
    cdef public dict config
    cdef public dict books

    cdef public exchange_manager.ExchangeManager exchange_manager

    cdef public str name
    cdef public str exchange_id

    cdef public list currencies
    cdef public list pairs
    cdef public list time_frames
    cdef public list channels

    cdef public object exchange
    cdef public object client
    cdef public object logger
    cdef public object bot_mainloop

    cpdef object get_exchange_credentials(self)
    cpdef object get_book_instance(self, str symbol)

    cpdef void add_pairs(self, list pairs, bint watching_only=*)
    cpdef void initialize(self, list currencies=*, list pairs=*, list time_frames=*, list channels=*)

    cpdef bint _should_authenticate(self)

    cpdef int get_max_handled_pair_with_time_frame(self)

    cpdef object feed_to_exchange(self, object feed)
    cpdef str get_pair_from_exchange(self, str pair)
    cpdef str get_exchange_pair(self, str pair)
    cpdef void clear(self)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal

import octobot_trading.constants as constants
import octobot_trading.errors as errors
import octobot_trading.personal_data.orders.order_util as order_util
import octobot_commons.symbols as symbols


class TradePnl:
    def __init__(self, entries, closes):
        self.entries = entries
        self.closes = closes

    def get_entry_time(self) -> float:
        try:
            return min(
                entry.executed_time
                for entry in self.entries
            )
        except ValueError as err:
            raise errors.IncompletePNLError from err

    def get_close_time(self) -> float:
        try:
            return max(
                close.executed_time
                for close in self.closes
            )
        except ValueError as err:
            raise errors.IncompletePNLError from err

    def get_total_entry_quantity(self) -> decimal.Decimal:
        return sum(
            entry.executed_quantity
            for entry in self.entries
        ) or constants.ZERO

    def get_total_close_quantity(self) -> decimal.Decimal:
        return sum(
            close.executed_quantity
            for close in self.closes
        ) or constants.ZERO

    def get_entry_price(self) -> decimal.Decimal:
        try:
            return sum(
                entry.executed_price
                for entry in self.entries
            ) / len(self.entries)
        except ZeroDivisionError as err:
            raise errors.IncompletePNLError from err

    def get_close_price(self) -> decimal.Decimal:
        try:
            return sum(
                close.executed_price
                for close in self.closes
            ) / len(self.closes)
        except ZeroDivisionError as err:
            raise errors.IncompletePNLError from err

    def get_closed_entry_value(self) -> decimal.Decimal:
        return self.get_entry_price() * self.get_closed_pnl_quantity()

    def get_closed_close_value(self) -> decimal.Decimal:
        return self.get_close_price() * self.get_closed_pnl_quantity()

    def get_closed_pnl_quantity(self):
        return min(self.get_total_close_quantity(), self.get_total_entry_quantity())

    def _get_fees(self, trade) -> decimal.Decimal:
        if not trade.fee:
            return constants.ZERO
        symbol = symbols.parse_symbol(trade.symbol)
        # return fees denominated in quote
        fees = order_util.get_fees_for_currency(trade.fee, symbol.quote)
        return fees \
            + order_util.get_fees_for_currency(trade.fee, symbol.base) * trade.executed_price

    def get_total_paid_fees(self) -> decimal.Decimal:
        return sum(
            self._get_fees(trade)
            for trade in (*self.entries, *self.closes)
        ) or constants.ZERO

    def get_profits(self) -> (decimal.Decimal, decimal.Decimal):
        """
        :return: the pnl profits as flat value and percent
        """
        close_holdings = self.get_closed_close_value() - self.get_total_paid_fees()
        entry_holdings = self.get_closed_entry_value()
        return (
            close_holdings - entry_holdings,
            (close_holdings * constants.ONE_HUNDRED / entry_holdings) - constants.ONE_HUNDRED
        )



# pylint: disable=E0611
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_trading.exchanges as exchanges
import octobot_commons.enums as commons_enums
import pytest

from tests.exchanges import exchange_manager, DEFAULT_EXCHANGE_NAME

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


class MockedCCXTConnector(exchanges.CCXTConnector):
    @classmethod
    def get_name(cls):
        return DEFAULT_EXCHANGE_NAME


class MockedRestExchange(exchanges.DefaultRestExchange):
    DEFAULT_CONNECTOR_CLASS = MockedCCXTConnector

    @classmethod
    def get_exchange_connector_class(cls, exchange_manager):
        return cls.DEFAULT_CONNECTOR_CLASS


@pytest.fixture
def default_rest_exchange(exchange_manager):
    return MockedRestExchange(exchange_manager.config, exchange_manager)


async def test_start_request_data_and_stop(default_rest_exchange):
    await default_rest_exchange.initialize()
    symbol = "BTC/USDT"
    try:
        assert len(default_rest_exchange.symbols) > 10
        assert len(default_rest_exchange.time_frames) > 5
        market_status = default_rest_exchange.get_market_status(symbol)
        assert isinstance(market_status, dict)
        assert market_status
        ohlcv = await default_rest_exchange.get_symbol_prices(symbol, commons_enums.TimeFrames.ONE_HOUR)
        assert isinstance(ohlcv, list)
        assert len(ohlcv) > 50
        trades = await default_rest_exchange.get_recent_trades(symbol)
        assert isinstance(trades, list)
        assert len(trades) > 5
        ticker = await default_rest_exchange.get_price_ticker(symbol)
        assert isinstance(ticker, dict)
        assert ticker
        book = await default_rest_exchange.get_order_book(symbol)
        assert isinstance(book, dict)
        assert book
    finally:
        await default_rest_exchange.stop()



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
cimport octobot_trading.personal_data.portfolios.asset as asset_class


cdef class FutureAsset(asset_class.Asset):
    cdef public object unrealized_pnl
    cdef public object initial_margin
    cdef public object wallet_balance
    cdef public object position_margin
    cdef public object order_margin

    cpdef object update(self, object total=*, object available=*, object position_margin=*,
                      object unrealized_pnl=*, object initial_margin=*)
    cpdef object set(self, object total=*, object available=*, object initial_margin=*, object position_margin=*,
                   object unrealized_pnl=*, object order_margin=*, object margin_balance=*)
    cpdef object set_unrealized_pnl(self, object unrealized_pnl)  # needs object to forward exceptions
    cpdef object update_realised_pnl(self, object realized_pnl_update)  # needs object to forward exceptions

    cdef object _update_total(self)
    cdef object _update_available(self)



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
cimport octobot_trading.exchange_data.ohlcv.candles_manager as candles_manager
cimport numpy as np


cdef class PreloadedCandlesManager(candles_manager.CandlesManager):
    cpdef int get_preloaded_symbol_candles_count(self)
    cpdef np.ndarray get_preloaded_symbol_close_candles(self)
    cpdef np.ndarray get_preloaded_symbol_open_candles(self)
    cpdef np.ndarray get_preloaded_symbol_high_candles(self)
    cpdef np.ndarray get_preloaded_symbol_low_candles(self)
    cpdef np.ndarray get_preloaded_symbol_time_candles(self)
    cpdef np.ndarray get_preloaded_symbol_volume_candles(self)

    # private
    cdef int _get_candle_index(self, list candle)
    cdef np.ndarray _get_candle_values_array(self, list candles, int key)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import octobot_trading.modes.script_keywords.dsl as dsl
import octobot_trading.modes.script_keywords.basic_keywords.account_balance as account_balance
import octobot_trading.errors as trading_errors
import octobot_trading.enums as trading_enums


async def get_amount_from_input_amount(
    context=None,
    input_amount=None,
    side=trading_enums.TradeOrderSide.BUY.value,
    reduce_only=True,
    is_stop_order=False,
    use_total_holding=False,
    target_price=None,
):
    amount_type, amount_value = dsl.parse_quantity(input_amount)

    if amount_type is dsl.QuantityType.UNKNOWN or amount_value <= 0:
        raise trading_errors.InvalidArgumentError("amount cant be zero or negative")

    if amount_type is dsl.QuantityType.DELTA:
        # nothing to do
        pass
    elif amount_type is dsl.QuantityType.PERCENT:
        amount_value = await account_balance.total_account_balance(context) * amount_value / 100
    elif amount_type is dsl.QuantityType.AVAILABLE_PERCENT:
        amount_value = await account_balance.available_account_balance(context, side, reduce_only=reduce_only) \
                       * amount_value / 100
    elif amount_type is dsl.QuantityType.POSITION_PERCENT:
        raise NotImplementedError(amount_type)
    else:
        raise trading_errors.InvalidArgumentError("make sure to use a supported syntax for amount")
    return await account_balance.adapt_amount_to_holdings(context, amount_value, side,
                                                          use_total_holding, reduce_only, is_stop_order,
                                                          target_price=target_price)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_trading.personal_data as personal_data


def create_trade(trader, trade_id, is_closing_order, origin_order_id):
    trade = personal_data.Trade(trader)
    trade.trade_id = trade_id
    trade.is_closing_order = is_closing_order
    trade.origin_order_id = origin_order_id
    return trade


def create_executed_trade(trader, executed_time, executed_quantity, executed_price, symbol, fee):
    trade = personal_data.Trade(trader)
    trade.executed_time = executed_time
    trade.executed_quantity = executed_quantity
    trade.executed_price = executed_price
    trade.symbol = symbol
    trade.fee = fee
    return trade



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import octobot_commons.constants as commons_constants
import octobot_commons.symbols.symbol_util as symbol_util
import octobot_trading.constants as trading_constants
import octobot_trading.enums as trading_enums
import octobot_trading.errors as trading_errors
import octobot_trading.personal_data as trading_personal_data


async def total_account_balance(context, side=trading_enums.PositionSide.BOTH.value):
    current_price = \
        await trading_personal_data.get_up_to_date_price(context.exchange_manager,
                                                         symbol=context.symbol,
                                                         timeout=trading_constants.ORDER_DATA_FETCHING_TIMEOUT)
    if current_price == trading_constants.ZERO:
        raise trading_errors.InvalidArgumentError("Current asset price is 0")
    on_inverse_contract = False
    if context.exchange_manager.is_future:
        # for BTC/USD
        # on linear, return balance in currency (ex: BTC)
        # on inverse, return balance in market (ex: USD)
        on_inverse_contract = context.exchange_manager.exchange_personal_data.positions_manager.get_symbol_position(
            context.symbol, trading_enums.PositionSide(side)
        ).symbol_contract.is_inverse_contract()
    value = context.exchange_manager.exchange_personal_data.\
        portfolio_manager.portfolio_value_holder.portfolio_current_value
    base, quote = symbol_util.parse_symbol(context.symbol).base_and_quote()
    reference_market = context.exchange_manager.exchange_personal_data.portfolio_manager.reference_market
    if reference_market == quote:
        return value if on_inverse_contract else value / current_price
    if reference_market == base:
        return value * current_price if on_inverse_contract else value
    raise NotImplementedError(f"Impossible to compute total account balance for a symbol that doesn't contain "
                              f"the reference market. reference market: {reference_market}, symbol: {context.symbol}")


def _get_locked_amount_in_stop_orders(context, side):
    locked_amount = trading_constants.ZERO
    for order in context.exchange_manager.exchange_personal_data.orders_manager.get_open_orders(context.symbol):
        if isinstance(order, (trading_personal_data.StopLossOrder, trading_personal_data.StopLossLimitOrder)) and \
           order.side.value == side:
            locked_amount += order.origin_quantity
    return locked_amount


async def available_account_balance(context, side=trading_enums.TradeOrderSide.BUY.value, use_total_holding=False,
                                    is_stop_order=False, reduce_only=True, target_price=None):
    portfolio_type = commons_constants.PORTFOLIO_TOTAL if use_total_holding else commons_constants.PORTFOLIO_AVAILABLE
    current_symbol_holding, _, market_quantity, price, _ = \
        await trading_personal_data.get_pre_order_data(context.exchange_manager,
                                                       symbol=context.symbol,
                                                       timeout=trading_constants.ORDER_DATA_FETCHING_TIMEOUT,
                                                       portfolio_type=portfolio_type,
                                                       target_price=target_price)
    already_locked_amount = trading_constants.ZERO
    if use_total_holding and is_stop_order:
        already_locked_amount = _get_locked_amount_in_stop_orders(context, side)
    if context.exchange_manager.is_future:
        max_order_size, _ = trading_personal_data.get_futures_max_order_size(
            context.exchange_manager, context.symbol, trading_enums.TradeOrderSide(side), price, reduce_only,
            current_symbol_holding, market_quantity
        )
        return max_order_size
    return (market_quantity if side == trading_enums.TradeOrderSide.BUY.value else current_symbol_holding) \
        - already_locked_amount


async def adapt_amount_to_holdings(context, amount, side, use_total_holding, reduce_only,
                                   is_stop_order, target_price=None):
    available_acc_bal = await available_account_balance(context, side, use_total_holding=use_total_holding,
                                                        is_stop_order=is_stop_order, reduce_only=reduce_only,
                                                        target_price=target_price)
    if available_acc_bal > amount:
        return amount
    else:
        return available_acc_bal


def account_holdings(context, currency):
    return context.exchange_manager.exchange_personal_data.portfolio_manager.portfolio.\
        get_currency_portfolio(currency).total


async def get_order_size_portfolio_percent(context, order_amount, input_side, symbol):
    try:
        return await trading_personal_data.get_order_size_portfolio_percent(
            context.exchange_manager,
            order_amount,
            trading_enums.TradeOrderSide(input_side),
            symbol
        )
    except ValueError:
        raise trading_errors.InvalidArgumentError(f"Unknown side: {input_side}")



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

# return object when an exception might be thrown
cpdef tuple create_client(object exchange_class, str exchange_name, object exchange_manager, object logger,
                                   dict options, dict headers, dict additional_config,
                                   bint should_authenticate, object unauthenticated_exchange_fallback=*)
cpdef object get_unauthenticated_exchange(object exchange_class, dict options, dict headers, dict additional_config)
cpdef object set_sandbox_mode(object exchange_connector, bint is_sandboxed)
cpdef dict get_ccxt_client_login_options(object exchange_manager)
cpdef set get_symbols(object client)
cpdef set get_time_frames(object client)
cpdef set get_exchange_pair(object client, str pair)
cpdef str get_pair_cryptocurrency(object client, str pair)
cpdef object get_contract_size(object client, str pair)
cpdef object add_headers(object client, dict headers_dict)
cpdef object add_options(object client, dict options_dict)

cdef object _use_http_proxy_if_necessary(object client)
cdef dict _get_client_config(dict options, dict headers, dict additional_config,
                              object api_key=*, object secret=*, object password=*)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest
from octobot_commons.tests.test_config import load_test_config

from tests import event_loop
from octobot_trading.personal_data.orders import Order, parse_order_type
from octobot_trading.enums import TradeOrderSide, TradeOrderType, TraderOrderType
from octobot_trading.exchanges.exchange_manager import ExchangeManager
from octobot_trading.exchanges.traders.trader_simulator import TraderSimulator
from octobot_trading.api.exchange import cancel_ccxt_throttle_task

pytestmark = pytest.mark.asyncio


class TestOrderFactory:
    DEFAULT_SYMBOL = "BTC/USDT"
    EXCHANGE_MANAGER_CLASS_STRING = "binanceus"

    @staticmethod
    async def init_default():
        config = load_test_config()
        exchange_manager = ExchangeManager(config, TestOrderFactory.EXCHANGE_MANAGER_CLASS_STRING)
        await exchange_manager.initialize()

        trader = TraderSimulator(config, exchange_manager)
        await trader.initialize()

        return config, exchange_manager, trader

    @staticmethod
    async def stop(exchange_manager):
        cancel_ccxt_throttle_task()
        await exchange_manager.stop()

    async def test_parse_order_type(self):
        _, exchange_manager, trader_inst = await self.init_default()

        order_to_test = Order(trader_inst)
        assert order_to_test.simulated is True

        ccxt_order_buy_market = {
            "side": TradeOrderSide.BUY,
            "type": TradeOrderType.MARKET
        }

        order_to_test.update_from_raw(ccxt_order_buy_market)
        assert order_to_test.order_type == TraderOrderType.BUY_MARKET

        ccxt_order_buy_limit = {
            "side": TradeOrderSide.BUY,
            "type": TradeOrderType.LIMIT
        }
        assert parse_order_type(ccxt_order_buy_limit) == (TradeOrderSide.BUY, TraderOrderType.BUY_LIMIT)

        ccxt_order_sell_market = {
            "side": TradeOrderSide.SELL,
            "type": TradeOrderType.MARKET
        }
        assert parse_order_type(ccxt_order_sell_market) == (TradeOrderSide.SELL, TraderOrderType.SELL_MARKET)

        ccxt_order_sell_limit = {
            "side": TradeOrderSide.SELL,
            "type": TradeOrderType.LIMIT
        }
        assert parse_order_type(ccxt_order_sell_limit) == (TradeOrderSide.SELL, TraderOrderType.SELL_LIMIT)

        await self.stop(exchange_manager)



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

cpdef object create_trade_instance_from_raw(object, dict raw_order)

cpdef object create_trade_from_order(object order,
                                     object close_status= *,
                                     double canceled_time= *,
                                     double creation_time= *,
                                     double executed_time= *)

cpdef object create_trade_instance(object,
                                   object order_type,
                                   str symbol,
                                   object status= *,
                                   str order_id= *,
                                   object filled_price= *,
                                   object quantity_filled= *,
                                   object total_cost= *,
                                   double canceled_time= *,
                                   double creation_time= *,
                                   double executed_time= *)

cpdef object create_trade_from_dict(object trader, dict trade_dict)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
from random import randrange, uniform

from octobot_commons.enums import PriceIndexes
from octobot_trading.enums import ExchangeConstantsOrderColumns as ECOC
import octobot_trading.constants as constants

MAX_PRICE = 10e7
MAX_INT = 10e7
MAX_QUANTITY = 10e7
DECIMAL_MAX_QUANTITY = decimal.Decimal("10e7")
MAX_FUNDING_RATE = 100
MAX_TIMESTAMP = 2000000000


def random_timestamp(min_value: int = 0, max_value=MAX_TIMESTAMP) -> float:
    return randrange(min_value, max_value)


def random_price(min_value: float = 0, max_value: float = None) -> float:
    return uniform(min_value, max_value if max_value is not None else MAX_PRICE)


def decimal_random_price(min_value=constants.ZERO,
                         max_value: decimal.Decimal = None) -> decimal.Decimal:
    return decimal.Decimal(str(uniform(float(min_value), float(max_value if max_value is not None else MAX_PRICE))))


def random_int(min_value: int = 0, max_value: int = None) -> int:
    return randrange(min_value, max_value if max_value is not None else MAX_INT)


def decimal_random_int(min_value: int = 0, max_value: int = None) -> decimal.Decimal:
    return decimal.Decimal(str(random_int(min_value=min_value, max_value=max_value)))


def random_prices(min_value: float = 0, count: int = 2) -> list:
    return [random_price(min_value=min_value) for _ in range(count)]


def random_quantity(min_value: float = 0, max_value: float = MAX_QUANTITY) -> float:
    return uniform(min_value, max_value)


def decimal_random_quantity(min_value=constants.ZERO,
                            max_value=DECIMAL_MAX_QUANTITY) -> decimal.Decimal:
    return decimal.Decimal(str(uniform(float(min_value), float(max_value))))


def random_quantities(min_value: float = 0, count: int = 2) -> list:
    return [random_quantity(min_value=min_value) for _ in range(count)]


def random_price_list(size: int = 2) -> list:
    return [random_price() for _ in range(size)]


def random_funding_rate(min_value: int = 0) -> float:
    return uniform(min_value, MAX_FUNDING_RATE)


def random_order_book_side(min_value: float = 0, count: int = 2) -> list:
    return [list(pair) for pair in zip(random_prices(min_value=min_value, count=count),
                                       random_quantities(min_value=min_value, count=count))]


def random_candle_tuple() -> tuple:
    return random_price(), random_price(), random_price(), random_price(), random_timestamp()


def random_candle() -> dict:
    return {
        PriceIndexes.IND_PRICE_CLOSE.value: random_price(),
        PriceIndexes.IND_PRICE_OPEN.value: random_price(),
        PriceIndexes.IND_PRICE_HIGH.value: random_price(),
        PriceIndexes.IND_PRICE_LOW.value: random_price(),
        PriceIndexes.IND_PRICE_VOL.value: random_quantity(),
        PriceIndexes.IND_PRICE_TIME.value: random_timestamp()
    }


def random_kline() -> list:
    kline = [0] * len(PriceIndexes)
    kline[PriceIndexes.IND_PRICE_CLOSE.value] = random_price()
    kline[PriceIndexes.IND_PRICE_OPEN.value] = random_price()
    kline[PriceIndexes.IND_PRICE_HIGH.value] = random_price()
    kline[PriceIndexes.IND_PRICE_LOW.value] = random_price()
    kline[PriceIndexes.IND_PRICE_VOL.value] = random_quantity()
    kline[PriceIndexes.IND_PRICE_TIME.value] = random_timestamp()
    return kline


def random_recent_trade(price=None, timestamp: float = None) -> dict:
    return {
        ECOC.PRICE.value: float(price) if price is not None else random_price(),
        ECOC.AMOUNT.value: random_quantity(),
        ECOC.COST.value: random_quantity(),
        ECOC.TIMESTAMP.value: timestamp if timestamp is not None else random_quantity(),
    }


def decimal_random_recent_trade(price=None, timestamp: float = None) -> dict:
    return {
        ECOC.PRICE.value: price if price is not None else decimal_random_price(),
        ECOC.AMOUNT.value: decimal_random_quantity(),
        ECOC.COST.value: decimal_random_quantity(),
        ECOC.TIMESTAMP.value: timestamp if timestamp is not None else decimal_random_quantity(),
    }


def random_recent_trades(count: int = 2) -> list:
    return [random_recent_trade() for _ in range(count)]



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

cimport octobot_trading.personal_data.orders.order as order_class

cdef class Portfolio:
    cdef object logger # Logger
    cdef public object lock # asyncio.Lock

    cdef public dict portfolio

    cdef str _exchange_name

    cdef bint _is_simulated

    # public methods
    cpdef void reset(self)
    # return object to ensure PortfolioNegativeValueError forwarding
    # cpdef object update_portfolio_from_balance(self, dict balance, bint force_replace=*) can't be cythonized for now
    cpdef object get_currency_portfolio(self, str currency)
    cpdef object update_portfolio_from_filled_order(self, order_class.Order order)
    cpdef object update_portfolio_from_withdrawal(self, object amount, str currency)
    cpdef object update_portfolio_available(self, order_class.Order order, bint is_new_order= *)
    # cpdef dict get_portfolio_from_amount_dict(self, dict amount_dict) can't be cythonized for now
    cpdef void reset_portfolio_available(self, str reset_currency= *, object reset_quantity= *)
    cpdef void log_portfolio_update_from_order(self, order_class.Order order)
    cpdef void log_portfolio_update_from_withdrawal(self, object amount, str currency)

    # abstract methods
    # return object to ensure PortfolioNegativeValueError forwarding
    cpdef object update_portfolio_data_from_order(self, order_class.Order order)
    cpdef object update_portfolio_available_from_order(self, order_class.Order order, bint is_new_order=*)
    cpdef object update_portfolio_data_from_withdrawal(self, object amount, str currency)
    cpdef object create_currency_asset(self, str currency, object available=*, object total=*)

    # private methods
    # return object to ensure PortfolioNegativeValueError forwarding
    cdef object _update_portfolio_data(self, str currency, object total_value=*, object available_value=*,
                                       bint replace_value=*)
    cdef object _parse_raw_currency_asset(self, str currency, dict raw_currency_balance)
    cdef bint _update_raw_currency_asset(self, str currency, dict raw_currency_balance)
    cdef void _reset_all_portfolio_available(self)
    cdef object _reset_currency_portfolio_available(self, str currency_to_reset, object reset_quantity)

cdef bint _should_reduce_available_assets_on_fill(order_class.Order order)
cdef bint _should_update_available(order_class.Order order)
cdef tuple _parse_raw_currency_balance(dict raw_currency_balance)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio

import octobot_trading.constants
import octobot_trading.enums as enums
import octobot_trading.errors
import octobot_trading.exchange_channel as exchange_channel
import octobot_trading.personal_data.state as state_class


class PositionState(state_class.State):
    def __init__(self, position, is_from_exchange_data):
        super().__init__(is_from_exchange_data)

        # related position
        self.position = position

        self._has_state_changed = asyncio.Event()

    def has_to_be_async_synchronized(self):
        return False

    def is_active(self) -> bool:
        """
        :return: True if the Position has a non-zero size
        """
        return False

    def is_liquidated(self) -> bool:
        """
        :return: True if the instance is considered as liquidated
        """
        return False

    def log_event_message(self, state_message, error=None):
        """
        Log a position state event
        """
        if state_message is enums.StatesMessages.ALREADY_SYNCHRONIZING:
            self.get_logger().debug(f"Trying to update a refreshing state for position: {self.position}")
        elif state_message is enums.StatesMessages.SYNCHRONIZING_ERROR:
            self.get_logger().exception(error, True, f"Error when synchronizing position {self.position}: {error}")
        else:
            self.get_logger().info(f"{self.position} {state_message.value} on "
                                   f"{self.position.exchange_manager.exchange_name}")

    async def _synchronize_with_exchange(self, force_synchronization: bool = False) -> None:
        """
        Ask PositionsChannel Internal producer to refresh the position from the exchange
        :param force_synchronization: When True, for the update of the position from the exchange
        :return: the result of PositionsProducer.update_position_from_exchange()
        """
        return (await exchange_channel.get_chan(octobot_trading.constants.POSITIONS_CHANNEL,
                                                self.position.exchange_manager.id).get_internal_producer().
                update_position_from_exchange(position=self.position,
                                              wait_for_refresh=True,
                                              force_job_execution=force_synchronization))

    def set_is_changing_state(self):
        if not self._has_state_changed.is_set():
            self._has_state_changed.set()

    def __del__(self):
        """
        Call set_is_changing_state in case this state has not been updated
        """
        self.set_is_changing_state()
        super().__del__()

    async def wait_for_next_state(self, timeout) -> None:
        if self._has_state_changed.is_set():
            return
        await asyncio.wait_for(self._has_state_changed.wait(), timeout=timeout)

    def clear(self):
        """
        Clear references
        """
        self.position = None



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest
import mock

import octobot_trading.exchanges as exchanges
import octobot_trading.enums as enums
import octobot_commons.tests.test_config as test_config

from tests import event_loop


EXCHANGE_NAME = "binanceus"


@pytest.fixture
def abstract_exchange():
    config = test_config.load_test_config()
    return exchanges.AbstractExchange(config, exchanges.ExchangeManager(config, EXCHANGE_NAME))

@pytest.mark.asyncio
async def test_log_order_creation_error(abstract_exchange):
    logger_mock = mock.Mock()
    abstract_exchange.logger = logger_mock
    error = RuntimeError()
    abstract_exchange.log_order_creation_error(error, enums.TraderOrderType.BUY_MARKET, "BTC/USD",
                                               1.0542454, 100.114, 100.114)
    logger_mock.error.assert_called_once()
    assert all(f"{e}" in logger_mock.mock_calls[0].args[0] for e in (enums.TraderOrderType.BUY_MARKET, "BTC/USD",
                                                                     1.0542454, 100.114, 100.114))
    logger_mock.reset_mock()
    abstract_exchange.log_order_creation_error(error, enums.TraderOrderType.BUY_MARKET, "BTC/USD", 0, None, None)
    logger_mock.error.assert_called_once()
    assert all(f"{e}" in logger_mock.mock_calls[0].args[0] for e in (enums.TraderOrderType.BUY_MARKET, "BTC/USD",
                                                                     0, None, None))
    logger_mock.reset_mock()


def test_supports_bundled_order_on_order_creation(abstract_exchange):
    order_mock = mock.Mock()
    order_mock.order_type = enums.TraderOrderType.SELL_MARKET
    assert abstract_exchange.supports_bundled_order_on_order_creation(order_mock, enums.TraderOrderType.STOP_LOSS) \
           is False
    abstract_exchange.SUPPORTED_BUNDLED_ORDERS[enums.TraderOrderType.SELL_MARKET] = [enums.TraderOrderType.SELL_MARKET,
                                                                                     enums.TraderOrderType.BUY_LIMIT]
    assert abstract_exchange.supports_bundled_order_on_order_creation(order_mock, enums.TraderOrderType.STOP_LOSS) \
           is False
    assert abstract_exchange.supports_bundled_order_on_order_creation(order_mock, enums.TraderOrderType.BUY_LIMIT) \
           is True


def test_get_order_additional_params(abstract_exchange):
    assert abstract_exchange.get_order_additional_params(None) == {}


def test_get_bundled_order_parameters(abstract_exchange):
    with pytest.raises(NotImplementedError):
        abstract_exchange.get_bundled_order_parameters(None)



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

cpdef object get_minimal_order_amount(dict symbol_market)
cpdef object decimal_adapt_price(dict symbol_market, object price, bint truncate=*)
cpdef object decimal_adapt_quantity(dict symbol_market, object quantity, bint truncate=*)
cpdef list decimal_adapt_order_quantity_because_quantity(object limiting_value, object max_value, object quantity_to_adapt, object price, dict symbol_market)
cpdef list decimal_adapt_order_quantity_because_price(object limiting_value, object max_value, object price, dict symbol_market)
cpdef list decimal_split_orders(object total_order_price, object max_cost, object valid_quantity, object max_quantity, object price, object quantity, dict symbol_market)
cpdef decimal_check_and_adapt_order_details_if_necessary(object quantity, object price, dict symbol_market, object fixed_symbol_data=*, bint truncate=*)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import mock
import pytest

import octobot_trading.modes.script_keywords as script_keywords


def test_parse_quantity_types():
    assert script_keywords.parse_quantity(None) == (script_keywords.QuantityType.DELTA, None)
    assert script_keywords.parse_quantity(10) == (script_keywords.QuantityType.DELTA, decimal.Decimal(10))
    assert script_keywords.parse_quantity(-10) == (script_keywords.QuantityType.DELTA, decimal.Decimal(-10))
    assert script_keywords.parse_quantity(1.366666663347877) == (
        script_keywords.QuantityType.DELTA, decimal.Decimal("1.366666663347877"))
    assert script_keywords.parse_quantity("-10") == (script_keywords.QuantityType.DELTA, decimal.Decimal(-10))

    assert script_keywords.parse_quantity("%") == (script_keywords.QuantityType.PERCENT, None)
    assert script_keywords.parse_quantity("99.5%") == (
        script_keywords.QuantityType.PERCENT, decimal.Decimal("99.5"))
    assert script_keywords.parse_quantity("-0.11%") == (
        script_keywords.QuantityType.PERCENT, decimal.Decimal("-0.11"))

    assert script_keywords.parse_quantity("a%") == (script_keywords.QuantityType.AVAILABLE_PERCENT, None)
    assert script_keywords.parse_quantity("-0.11a%") == (
        script_keywords.QuantityType.AVAILABLE_PERCENT, decimal.Decimal("-0.11"))
    assert script_keywords.parse_quantity("%a-0.11") == (
        script_keywords.QuantityType.AVAILABLE_PERCENT, decimal.Decimal("-0.11"))

    assert script_keywords.parse_quantity("a") == (script_keywords.QuantityType.AVAILABLE, None)
    assert script_keywords.parse_quantity("-0.11a") == (
        script_keywords.QuantityType.AVAILABLE, decimal.Decimal("-0.11"))
    assert script_keywords.parse_quantity("a-0.11") == (
        script_keywords.QuantityType.AVAILABLE, decimal.Decimal("-0.11"))

    assert script_keywords.parse_quantity("e%") == (script_keywords.QuantityType.ENTRY_PERCENT, None)
    assert script_keywords.parse_quantity("-0.11e%") == (
        script_keywords.QuantityType.ENTRY_PERCENT, decimal.Decimal("-0.11"))
    assert script_keywords.parse_quantity("-0.11%e") == (
        script_keywords.QuantityType.ENTRY_PERCENT, decimal.Decimal("-0.11"))

    assert script_keywords.parse_quantity("e") == (script_keywords.QuantityType.ENTRY, None)
    assert script_keywords.parse_quantity("-0.11e") == (
        script_keywords.QuantityType.ENTRY, decimal.Decimal("-0.11"))
    assert script_keywords.parse_quantity("e-0.11") == (
        script_keywords.QuantityType.ENTRY, decimal.Decimal("-0.11"))

    assert script_keywords.parse_quantity("p%") == (script_keywords.QuantityType.POSITION_PERCENT, None)
    assert script_keywords.parse_quantity("-0.11p%") == (
        script_keywords.QuantityType.POSITION_PERCENT, decimal.Decimal("-0.11"))
    assert script_keywords.parse_quantity("%p-0.11") == (
        script_keywords.QuantityType.POSITION_PERCENT, decimal.Decimal("-0.11"))

    assert script_keywords.parse_quantity("p") == (script_keywords.QuantityType.POSITION, None)
    assert script_keywords.parse_quantity("-0.11p") == (
        script_keywords.QuantityType.POSITION, decimal.Decimal("-0.11"))
    assert script_keywords.parse_quantity("p-0.11") == (
        script_keywords.QuantityType.POSITION, decimal.Decimal("-0.11"))

    assert script_keywords.parse_quantity("@") == (script_keywords.QuantityType.FLAT, None)
    assert script_keywords.parse_quantity("-0.11@") == (script_keywords.QuantityType.FLAT, decimal.Decimal("-0.11"))
    assert script_keywords.parse_quantity("@-0.11") == (script_keywords.QuantityType.FLAT, decimal.Decimal("-0.11"))

    assert script_keywords.parse_quantity("wyz-0.11") == (script_keywords.QuantityType.UNKNOWN, None)
    assert script_keywords.parse_quantity("wyz12") == (script_keywords.QuantityType.UNKNOWN, None)
    assert script_keywords.parse_quantity("wyz") == (script_keywords.QuantityType.UNKNOWN, None)


def test_parse_quantity_edge_numbers():
    assert script_keywords.parse_quantity(0) == (script_keywords.QuantityType.DELTA, decimal.Decimal(0))
    assert script_keywords.parse_quantity(0.000000001) == (
        script_keywords.QuantityType.DELTA, decimal.Decimal("0.000000001"))
    assert script_keywords.parse_quantity(100000000000000000000) \
           == (script_keywords.QuantityType.DELTA, decimal.Decimal("100000000000000000000"))
    assert script_keywords.parse_quantity("-1e-09e%") == (
        script_keywords.QuantityType.ENTRY_PERCENT, decimal.Decimal("-1e-09"))
    assert script_keywords.parse_quantity("1e09%e") == (
        script_keywords.QuantityType.ENTRY_PERCENT, decimal.Decimal("1e09"))
    assert script_keywords.parse_quantity("1e9e") == (script_keywords.QuantityType.ENTRY, decimal.Decimal("1e9"))
    with mock.patch.object(script_keywords.QuantityType, "parse",
                           mock.Mock(return_value=(script_keywords.QuantityType.ENTRY, "e"))):
        with pytest.raises(RuntimeError):
            script_keywords.parse_quantity("0.000000001aa")



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
cimport octobot_trading.personal_data.orders.order as order_class
cimport octobot_trading.exchanges as exchanges
cimport octobot_trading.util as util


cdef class OrdersManager(util.Initializable):
    cdef object logger

    cdef public bint orders_initialized

    cdef exchanges.Trader trader

    cdef public object orders
    cdef public dict order_groups
    cdef public list pending_creation_orders
    cdef public bint are_exchange_orders_initialized

    cdef void _reset_orders(self)
    cdef void _check_orders_size(self)
    cdef void _remove_oldest_orders(self, int nb_to_remove)
    cdef list _select_orders(self, object state=*, str symbol=*, int since=*, int until=*, int limit=*, str tag=*)
    cdef object _get_pending_order(self, object created_order, bint should_pop)
    cdef void _add_order(self, str order_id, object order)

    cpdef order_class.Order get_order(self, str order_id)
    cpdef void register_pending_creation_order(self, object pending_order)
    cpdef bint has_order(self, str order_id)
    cpdef void remove_order_instance(self, order_class.Order order)
    cpdef void replace_order(self, str previous_id, order_class.Order order)
    cpdef list get_all_orders(self, str symbol=*, int since=*, int until=*, int limit=*, str tag=*)
    cpdef list get_open_orders(self, str symbol=*, int since=*, int until=*, int limit=*, str tag=*)
    cpdef list get_pending_cancel_orders(self, str symbol=*, int since=*, int until=*, int limit=*, str tag=*)
    cpdef list get_closed_orders(self, str symbol=*, int since=*, int until=*, int limit=*, str tag=*)
    cpdef list get_order_from_group(self, str group_name)
    cpdef object get_or_create_group(self, object group_type, str group_name)
    cpdef void clear(self)



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
cimport octobot_trading.modes.channel as modes_channel

cdef class AbstractTradingModeProducer(modes_channel.ModeChannelProducer):
    cdef public object trading_mode
    cdef public object config
    cdef public object exchange_manager
    cdef public object final_eval
    cdef public object state
    cdef public object matrix_consumer

    cdef public str exchange_name

    cdef public int priority_level

    cdef public object _is_ready_to_trade

    cpdef void flush(self)
    cpdef bint is_cryptocurrency_wildcard(self)
    cpdef bint is_symbol_wildcard(self)
    cpdef bint is_time_frame_wildcard(self)
    cpdef object get_callback(self, str chan_name)
    cpdef list get_channels_registration(self)
    cpdef object get_trigger_time_frames(self)



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

cdef class Asset:
    cdef public str name

    cdef public object available  # Decimal
    cdef public object total  # Decimal

    cpdef void restore_available(self)
    cpdef void reset(self)
    cpdef dict to_dict(self)

    cpdef object _ensure_update_validity(self, object origin_quantity, object update_quantity)
    cpdef object _ensure_not_negative(self, object new_value, object replacement_value=*)
    cpdef object restore_unavailable_from_other(self, object other_asset)
    cpdef object _specific_restore_unavailable_from_other(self, object other_asset)



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
cimport octobot_trading.util as util


cdef class TransactionsManager(util.Initializable):
    cdef object logger

    cdef public object transactions

    cdef void _check_transactions_size(self)
    cdef void _reset_transactions(self)
    cdef void _remove_oldest_transactions(self, int nb_to_remove)

    cpdef object get_transaction(self, str transaction_id)
    cpdef object update_transaction_id(self, str transaction_id, str new_transaction_id, bint replace_if_exists=*)  # needs object to forward exceptions
    cpdef object insert_transaction_instance(self, object transaction, bint replace_if_exists=*)  # needs object to forward exceptions
    cpdef void clear(self)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_trading.personal_data as personal_data
import octobot_trading.enums as enums
import octobot_trading.constants as constants


def create_order_from_raw(trader, raw_order):
    _, order_type = personal_data.parse_order_type(raw_order)
    return create_order_from_type(trader, order_type)


def create_order_instance_from_raw(trader, raw_order, force_open_or_pending_creation=False):
    order = create_order_from_raw(trader, raw_order)
    order.update_from_raw(raw_order)
    if force_open_or_pending_creation \
            and order.status not in (enums.OrderStatus.OPEN, enums.OrderStatus.PENDING_CREATION):
        order.status = enums.OrderStatus.OPEN
    return order


def create_order_from_type(trader, order_type, side=None):
    if side is None:
        return personal_data.TraderOrderTypeClasses[order_type](trader)
    return personal_data.TraderOrderTypeClasses[order_type](trader, side=side)


def create_order_instance(trader,
                          order_type,
                          symbol,
                          current_price,
                          quantity,
                          price=constants.ZERO,
                          stop_price=constants.ZERO,
                          status=enums.OrderStatus.OPEN,
                          order_id=None,
                          filled_price=constants.ZERO,
                          average_price=constants.ZERO,
                          quantity_filled=constants.ZERO,
                          total_cost=constants.ZERO,
                          timestamp=0,
                          side=None,
                          fees_currency_side=None,
                          group=None,
                          tag=None,
                          reduce_only=None,
                          quantity_currency=None,
                          close_position=False,
                          exchange_creation_params=None,
                          associated_entry_id=None):
    order = create_order_from_type(trader=trader,
                                   order_type=order_type,
                                   side=side)
    order.update(order_type=order_type,
                 symbol=symbol,
                 current_price=current_price,
                 quantity=quantity,
                 price=price,
                 stop_price=stop_price,
                 order_id=trader.parse_order_id(order_id),
                 timestamp=timestamp,
                 status=status,
                 filled_price=filled_price,
                 average_price=average_price,
                 quantity_filled=quantity_filled,
                 fee=None,
                 total_cost=total_cost,
                 fees_currency_side=fees_currency_side,
                 group=group,
                 tag=tag,
                 reduce_only=reduce_only,
                 quantity_currency=quantity_currency,
                 close_position=close_position,
                 exchange_creation_params=exchange_creation_params,
                 associated_entry_id=associated_entry_id)
    order.ensure_order_id()
    return order



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import math
import decimal

import octobot_trading.constants as constants
import octobot_trading.errors as errors
import octobot_trading.exchanges as exchanges
import octobot_trading.personal_data as personal_data
from octobot_trading.enums import ExchangeConstantsMarketStatusColumns as Ecmsc

DECIMAL_SCIENTIFIC_NOTATION_EXP = "E-"


def get_minimal_order_amount(symbol_market):
    try:
        min_limit = symbol_market[Ecmsc.LIMITS.value][Ecmsc.LIMITS_AMOUNT.value].get(Ecmsc.LIMITS_AMOUNT_MIN.value,
                                                                                     None)
        if min_limit is not None:
            return decimal.Decimal(str(min_limit))
        min_precision = symbol_market[Ecmsc.PRECISION.value][Ecmsc.PRECISION_AMOUNT.value]
        return decimal.Decimal(f"1e-{min_precision}")
    except KeyError:
        raise errors.NotSupported("Impossible to get the minimal order size for the this exchange")


def decimal_adapt_price(symbol_market, price, truncate=True):
    maximal_price_digits = symbol_market[Ecmsc.PRECISION.value].get(
                                                Ecmsc.PRECISION_PRICE.value,
                                                constants.CURRENCY_DEFAULT_MAX_PRICE_DIGITS)
    return decimal_trunc_with_n_decimal_digits(price, maximal_price_digits, truncate)


def decimal_adapt_quantity(symbol_market, quantity, truncate=True):
    maximal_volume_digits = symbol_market[Ecmsc.PRECISION.value].get(
                                                 Ecmsc.PRECISION_AMOUNT.value, 0)
    return decimal_trunc_with_n_decimal_digits(quantity, maximal_volume_digits, truncate)


def _has_more_than_x_digits(value, digits):
    str_val = str(value)
    if len(str_val.split(".")[-1]) > digits:
        return True
    if DECIMAL_SCIENTIFIC_NOTATION_EXP in str_val and int(str_val.split(DECIMAL_SCIENTIFIC_NOTATION_EXP)[1]) > digits:
        return True
    return False


def decimal_trunc_with_n_decimal_digits(value, digits, truncate=True):  # TODO migrate to commons
    try:
        # decimal.Decimal can add unnecessary complexity in numbers, only use it when necessary
        if _has_more_than_x_digits(value, digits):
            if digits > constants.ZERO:
                return value.quantize(decimal.Decimal(f".{'0' * int(digits)}"),
                                      rounding=decimal.ROUND_DOWN if truncate else decimal.ROUND_UP)
            else:
                return value // constants.ONE
        return value
    except (ValueError, decimal.InvalidOperation):
        return value


def decimal_adapt_order_quantity_because_quantity(limiting_value, max_value, quantity_to_adapt, price, symbol_market):
    orders = []
    nb_full_orders = limiting_value // max_value
    rest_order_quantity = limiting_value % max_value
    after_rest_quantity_to_adapt = quantity_to_adapt

    if rest_order_quantity > constants.ZERO:
        after_rest_quantity_to_adapt -= rest_order_quantity
        valid_last_order_quantity = decimal_adapt_quantity(symbol_market, rest_order_quantity)
        orders.append((valid_last_order_quantity, price))

    other_orders_quantity = (after_rest_quantity_to_adapt + max_value) / (nb_full_orders + constants.ONE)
    valid_other_orders_quantity = decimal_adapt_quantity(symbol_market, other_orders_quantity)
    orders += [(valid_other_orders_quantity, price)] * int(nb_full_orders)
    return orders


def decimal_adapt_order_quantity_because_price(limiting_value, max_value, price, symbol_market):
    orders = []
    nb_full_orders = limiting_value // max_value
    rest_order_cost = limiting_value % max_value
    if rest_order_cost > constants.ZERO:
        valid_last_order_quantity = decimal_adapt_quantity(symbol_market, rest_order_cost / price)
        orders.append((valid_last_order_quantity, price))

    other_orders_quantity = max_value / price
    valid_other_orders_quantity = decimal_adapt_quantity(symbol_market, other_orders_quantity)
    orders += [(valid_other_orders_quantity, price)] * int(nb_full_orders)
    return orders


def decimal_split_orders(total_order_price, max_cost, valid_quantity, max_quantity, price, quantity, symbol_market):
    """
    Splits too big orders into multiple ones according to the max_cost and max_quantity
    :param total_order_price:
    :param max_cost:
    :param valid_quantity:
    :param max_quantity:
    :param price:
    :param quantity:
    :param symbol_market:
    :return:
    """
    if max_cost is None and max_quantity is None:
        raise RuntimeError("Impossible to split orders with max_cost and max_quantity undefined.")
    nb_orders_according_to_cost = None
    nb_orders_according_to_quantity = None
    if max_cost:
        nb_orders_according_to_cost = total_order_price / max_cost

    if max_quantity:
        nb_orders_according_to_quantity = valid_quantity / max_quantity

    if nb_orders_according_to_cost is None:
        # can only split using quantity
        return decimal_adapt_order_quantity_because_quantity(valid_quantity, max_quantity,
                                                             quantity, price, symbol_market)
    elif nb_orders_according_to_quantity is None:
        # can only split using price
        return decimal_adapt_order_quantity_because_price(total_order_price, max_cost, price, symbol_market)
    else:
        if nb_orders_according_to_cost > nb_orders_according_to_quantity:
            return decimal_adapt_order_quantity_because_price(total_order_price, max_cost, price, symbol_market)
        return decimal_adapt_order_quantity_because_quantity(valid_quantity, max_quantity,
                                                             quantity, price, symbol_market)


def decimal_check_and_adapt_order_details_if_necessary(quantity, price, symbol_market, fixed_symbol_data=False,
                                                       truncate=True):
    """
    Checks if order attributes are valid and try to fix it if not
    :param quantity:
    :param price:
    :param symbol_market:
    :param fixed_symbol_data:
    :param truncate:
    :return:
    """
    if quantity.is_nan() or price.is_nan() or price == constants.ZERO:
        return []

    symbol_market_limits = symbol_market[Ecmsc.LIMITS.value]

    limit_amount = symbol_market_limits[Ecmsc.LIMITS_AMOUNT.value]
    limit_cost = symbol_market_limits[Ecmsc.LIMITS_COST.value]
    limit_price = symbol_market_limits[Ecmsc.LIMITS_PRICE.value]

    # adapt digits if necessary
    valid_quantity = decimal_adapt_quantity(symbol_market, quantity, truncate)
    valid_price = decimal_adapt_price(symbol_market, price, truncate)

    # case 1: try with data directly from exchange
    if personal_data.is_valid(limit_amount, Ecmsc.LIMITS_AMOUNT_MIN.value):
        min_quantity = decimal.Decimal(str(limit_amount.get(Ecmsc.LIMITS_AMOUNT_MIN.value, math.nan)))
        max_quantity = None
        # not all symbol data have a max quantity
        if personal_data.is_valid(limit_amount, Ecmsc.LIMITS_AMOUNT_MAX.value):
            max_quantity = decimal.Decimal(str(limit_amount.get(Ecmsc.LIMITS_AMOUNT_MAX.value, math.nan)))

        total_order_price = valid_quantity * valid_price

        if valid_quantity < min_quantity:
            # invalid order
            return []

        # case 1.1: use only quantity and cost
        if personal_data.is_valid(limit_cost, Ecmsc.LIMITS_COST_MIN.value):
            min_cost = decimal.Decimal(str(limit_cost.get(Ecmsc.LIMITS_COST_MIN.value, math.nan)))
            max_cost = None
            # not all symbol data have a max cost
            if personal_data.is_valid(limit_cost, Ecmsc.LIMITS_COST_MAX.value):
                max_cost = decimal.Decimal(str(limit_cost.get(Ecmsc.LIMITS_COST_MAX.value, math.nan)))

            # check total_order_price not < min_cost
            if not personal_data.check_cost(float(total_order_price), min_cost):
                return []

            # check total_order_price not > max_cost and valid_quantity not > max_quantity
            elif (max_cost is not None and total_order_price > max_cost) or \
                    (max_quantity is not None and valid_quantity > max_quantity):
                # split quantity into smaller orders
                return decimal_split_orders(total_order_price, max_cost, valid_quantity,
                                            max_quantity, valid_price, quantity, symbol_market)

            else:
                # valid order that can be handled by the exchange
                return [(valid_quantity, valid_price)]

        # case 1.2: use only quantity and price
        elif personal_data.is_valid(limit_price, Ecmsc.LIMITS_PRICE_MIN.value):
            min_price = decimal.Decimal(str(limit_price.get(Ecmsc.LIMITS_PRICE_MIN.value, math.nan)))
            max_price = None
            # not all symbol data have a max price
            if personal_data.is_valid(limit_price, Ecmsc.LIMITS_PRICE_MAX.value):
                max_price = decimal.Decimal(str(limit_price.get(Ecmsc.LIMITS_PRICE_MAX.value, math.nan)))

            if (max_price is not None and (max_price <= valid_price)) or valid_price <= min_price:
                # invalid order
                return []

            # check total_order_price not > max_cost and valid_quantity not > max_quantity
            elif max_quantity is not None and valid_quantity > max_quantity:
                # split quantity into smaller orders
                return decimal_adapt_order_quantity_because_quantity(valid_quantity, max_quantity,
                                                                     quantity, valid_price, symbol_market)
            else:
                # valid order that can be handled wy the exchange
                return [(valid_quantity, valid_price)]

    if not fixed_symbol_data:
        # case 2: try fixing data from exchanges
        fixed_data = exchanges.ExchangeMarketStatusFixer(symbol_market, float(price)).market_status
        return decimal_check_and_adapt_order_details_if_necessary(quantity, price, fixed_data,
                                                                  fixed_symbol_data=True, truncate=truncate)
    else:
        # impossible to check if order is valid: try anyway, the exchange will tell
        return [(valid_quantity, valid_price)]


def decimal_add_dusts_to_quantity_if_necessary(quantity, price, symbol_market, current_symbol_holding):
    """
    Adds remaining quantity to the order if the remaining quantity is too small
    :param quantity:
    :param price:
    :param symbol_market:
    :param current_symbol_holding:
    :return:
    """
    if price == constants.ZERO:
        return quantity

    remaining_portfolio_amount = decimal.Decimal("{1:.{0}f}".format(constants.CURRENCY_DEFAULT_MAX_PRICE_DIGITS,
                                                                    current_symbol_holding - quantity))
    remaining_max_total_order_price = remaining_portfolio_amount * price

    symbol_market_limits = symbol_market[Ecmsc.LIMITS.value]

    limit_amount = symbol_market_limits[Ecmsc.LIMITS_AMOUNT.value]
    limit_cost = symbol_market_limits[Ecmsc.LIMITS_COST.value]

    if not (personal_data.is_valid(limit_amount, Ecmsc.LIMITS_AMOUNT_MIN.value) and
            personal_data.is_valid(limit_cost, Ecmsc.LIMITS_COST_MIN.value)):
        fixed_market_status = exchanges.ExchangeMarketStatusFixer(symbol_market, float(price)).market_status
        limit_amount = fixed_market_status[Ecmsc.LIMITS.value][Ecmsc.LIMITS_AMOUNT.value]
        limit_cost = fixed_market_status[Ecmsc.LIMITS.value][Ecmsc.LIMITS_COST.value]

    min_quantity = limit_amount.get(Ecmsc.LIMITS_AMOUNT_MIN.value, math.nan)
    min_cost = limit_cost.get(Ecmsc.LIMITS_COST_MIN.value, math.nan)

    # check with 40% more than remaining total not to require huge market moves to sell this asset
    min_cost_to_consider = decimal.Decimal(str(min_cost)) * decimal.Decimal(str(1.4))
    min_quantity_to_consider = decimal.Decimal(str(min_quantity)) * decimal.Decimal(str(1.4))

    if remaining_max_total_order_price < min_cost_to_consider \
            or remaining_portfolio_amount < min_quantity_to_consider:
        return current_symbol_holding
    else:
        return quantity



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest

from tests import event_loop
from octobot_commons.tests.test_config import load_test_config
from octobot_commons.constants import CONFIG_CRYPTO_CURRENCIES
from octobot_trading.exchanges.exchange_manager import ExchangeManager
from octobot_trading.api.exchange import cancel_ccxt_throttle_task

pytestmark = pytest.mark.asyncio


class TestExchangeConfig:
    EXCHANGE_NAME = "binanceus"

    @staticmethod
    async def init_default(config=None):
        if not config:
            config = load_test_config()

        exchange_manager = ExchangeManager(config, TestExchangeConfig.EXCHANGE_NAME)

        await exchange_manager.initialize()
        return config, exchange_manager

    async def test_traded_pairs(self):
        config = load_test_config()
        config[CONFIG_CRYPTO_CURRENCIES] = {
            "Avalanche": {
                "pairs": ["AVAX/BTC"]
            },
            "Ethereum": {
                "enabled": True,
                "pairs": ["ETH/USDT"]
            },
            "Uniswap": {
                "enabled": False,
                "pairs": ["UNI/BTC"]
            }
        }

        _, exchange_manager = await self.init_default(config=config)

        assert exchange_manager.exchange_config.traded_cryptocurrencies == {
            "Ethereum": ["ETH/USDT"],
            "Avalanche": ["AVAX/BTC"]
        }
        all_pairs = sorted(["AVAX/BTC", "ETH/USDT", "UNI/BTC"])
        all_enabled_pairs = sorted(["AVAX/BTC", "ETH/USDT"])
        assert sorted(exchange_manager.exchange_config.traded_symbol_pairs) == all_enabled_pairs
        assert sorted(exchange_manager.exchange_config.all_config_symbol_pairs) == all_pairs
        cancel_ccxt_throttle_task()
        await exchange_manager.stop()

    async def test_traded_pairs_with_wildcard(self):
        config = load_test_config()
        config[CONFIG_CRYPTO_CURRENCIES] = {
            "Bitcoin": {
                "pairs": ["*"],
                "quote": "BTC"
            },
            "Ethereum": {
                "enabled": False,
                "pairs": ["*"],
                "quote": "ETH"
            }
        }
        _, exchange_manager = await self.init_default(config=config)

        assert "UNI/BTC" in exchange_manager.exchange_config.traded_cryptocurrencies["Bitcoin"]
        assert "AVAX/BTC" in exchange_manager.exchange_config.traded_cryptocurrencies["Bitcoin"]
        assert "ADA/BTC" in exchange_manager.exchange_config.traded_cryptocurrencies["Bitcoin"]
        assert "MATIC/BTC" in exchange_manager.exchange_config.traded_cryptocurrencies["Bitcoin"]
        assert "ONT/BTC" not in exchange_manager.exchange_config.traded_cryptocurrencies["Bitcoin"]
        assert "BTC/USDT" not in exchange_manager.exchange_config.traded_cryptocurrencies["Bitcoin"]
        assert "ETH/USDT" not in exchange_manager.exchange_config.traded_cryptocurrencies["Bitcoin"]
        assert "AVAX/BNB" not in exchange_manager.exchange_config.traded_cryptocurrencies["Bitcoin"]
        assert "ETH/BTC" in exchange_manager.exchange_config.traded_symbol_pairs
        assert "ETH/BTC" in exchange_manager.exchange_config.all_config_symbol_pairs

        # disabled
        assert "Ethereum" not in exchange_manager.exchange_config.traded_cryptocurrencies
        assert "ADA/ETH" not in exchange_manager.exchange_config.traded_symbol_pairs

        cancel_ccxt_throttle_task()
        await exchange_manager.stop()

    async def test_traded_pairs_with_invalid_wildcard(self):
        config = load_test_config()

        # missing quote key
        config[CONFIG_CRYPTO_CURRENCIES] = {
            "Bitcoin": {
                "enabled": True,
                "pairs": ["*"],
                "quote": "BTC"
            },
            "Ethereum": {
                "pairs": ["*"],
            }
        }
        _, exchange_manager = await self.init_default(config=config)

        assert "TRX/BTC" in exchange_manager.exchange_config.traded_symbol_pairs
        assert "ADA/BTC" in exchange_manager.exchange_config.traded_symbol_pairs
        assert "BNB/BTC" in exchange_manager.exchange_config.all_config_symbol_pairs
        assert "ADA/BTC" in exchange_manager.exchange_config.all_config_symbol_pairs
        assert "Bitcoin" in exchange_manager.exchange_config.traded_cryptocurrencies

        # invalid ETH wildcard config
        assert "Ethereum" not in exchange_manager.exchange_config.traded_cryptocurrencies

        cancel_ccxt_throttle_task()
        await exchange_manager.stop()

    async def test_traded_pairs_with_add(self):
        config = load_test_config()
        config[CONFIG_CRYPTO_CURRENCIES] = {
            "Bitcoin": {
                "pairs": ["*"],
                "quote": "BTC",
                "add": ["BTC/USDT"]
            },
            "Ethereum": {
                "enabled": False,
                "pairs": ["*"],
                "quote": "ETH",
                "add": ["ETH/USDT"]
            }
        }

        _, exchange_manager = await self.init_default(config=config)

        assert "UNI/BTC" in exchange_manager.exchange_config.traded_cryptocurrencies["Bitcoin"]
        assert "AVAX/BTC" in exchange_manager.exchange_config.traded_cryptocurrencies["Bitcoin"]
        assert "ADA/BTC" in exchange_manager.exchange_config.traded_cryptocurrencies["Bitcoin"]
        assert "MATIC/BTC" in exchange_manager.exchange_config.traded_cryptocurrencies["Bitcoin"]
        assert "LINK/BTC" in exchange_manager.exchange_config.traded_cryptocurrencies["Bitcoin"]
        assert "ONT/BTC" not in exchange_manager.exchange_config.traded_cryptocurrencies["Bitcoin"]
        assert "BTC/USDT" in exchange_manager.exchange_config.traded_cryptocurrencies["Bitcoin"]
        assert "ETH/USDT" not in exchange_manager.exchange_config.traded_cryptocurrencies["Bitcoin"]
        assert "AVAX/BNB" not in exchange_manager.exchange_config.traded_cryptocurrencies["Bitcoin"]
        assert "BTC/USDT" in exchange_manager.exchange_config.traded_symbol_pairs
        assert "BTC/USDT" in exchange_manager.exchange_config.all_config_symbol_pairs

        # disabled
        assert "Ethereum" not in exchange_manager.exchange_config.traded_cryptocurrencies
        assert "ADA/ETH" not in exchange_manager.exchange_config.traded_symbol_pairs
        assert "ETH/USDT" not in exchange_manager.exchange_config.traded_symbol_pairs
        assert "ETH/USDT" in exchange_manager.exchange_config.all_config_symbol_pairs
        cancel_ccxt_throttle_task()
        await exchange_manager.stop()

    async def test_traded_pairs_with_redundancy(self):
        config = load_test_config()
        config[CONFIG_CRYPTO_CURRENCIES] = {
            "Binance Coin": {
                "pairs": [
                    "BNB/USDT"
                ]
            },
            "Binance USD": {
                "pairs": [
                    "BNB/BUSD"
                ]
            },
            "Bitcoin": {
                "enabled": True,
                "pairs": [
                    "BNB/BTC"
                ]
            },
            "Tether": {
                "enabled": True,
                "pairs": [
                    "BNB/USDT"
                ]
            }
        }

        _, exchange_manager = await self.init_default(config=config)

        assert exchange_manager.exchange_config.traded_cryptocurrencies["Binance Coin"] == ["BNB/USDT"]
        assert exchange_manager.exchange_config.traded_cryptocurrencies["Binance USD"] == ["BNB/BUSD"]
        assert exchange_manager.exchange_config.traded_cryptocurrencies["Bitcoin"] == ["BNB/BTC"]
        assert exchange_manager.exchange_config.traded_cryptocurrencies["Tether"] == ["BNB/USDT"]

        sorted_pairs_without_redundancy = sorted(["BNB/USDT", "BNB/BUSD", "BNB/BTC"])
        assert sorted(exchange_manager.exchange_config.traded_symbol_pairs) == sorted_pairs_without_redundancy
        assert sorted(exchange_manager.exchange_config.all_config_symbol_pairs) == sorted_pairs_without_redundancy

        cancel_ccxt_throttle_task()
        await exchange_manager.stop()



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import pytest

# avoid circular imports when launching tests from this folder
import octobot_trading.api  # TODO fix circular import when importing octobot_trading.exchange_data first

from octobot_trading.exchange_data.prices.price_events_manager import PriceEventsManager
from octobot_trading.exchange_data.prices.prices_manager import PricesManager
from octobot_trading.exchange_data.recent_trades.recent_trades_manager import RecentTradesManager


@pytest.fixture()
def price_events_manager(event_loop):
    return PriceEventsManager()


@pytest.fixture()
def prices_manager(event_loop, backtesting_exchange_manager):
    return PricesManager(backtesting_exchange_manager)


@pytest.fixture()
def recent_trades_manager(event_loop):
    return RecentTradesManager()



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_trading.personal_data.orders.states cimport close_order_state
from octobot_trading.personal_data.orders.states.close_order_state cimport (
    CloseOrderState
)

from octobot_trading.personal_data.orders.states cimport cancel_order_state
from octobot_trading.personal_data.orders.states.cancel_order_state cimport (
    CancelOrderState
)

from octobot_trading.personal_data.orders.states cimport open_order_state
from octobot_trading.personal_data.orders.states.open_order_state cimport (
    OpenOrderState
)

from octobot_trading.personal_data.orders.states cimport pending_creation_order_state
from octobot_trading.personal_data.orders.states.pending_creation_order_state cimport (
    PendingCreationOrderState
)

from octobot_trading.personal_data.orders.states cimport pending_creation_chained_order_state
from octobot_trading.personal_data.orders.states.pending_creation_chained_order_state cimport (
    PendingCreationChainedOrderState
)

from octobot_trading.personal_data.orders.states cimport fill_order_state
from octobot_trading.personal_data.orders.states.fill_order_state cimport (
    FillOrderState
)


__all__ = [
    "CloseOrderState",
    "CancelOrderState",
    "OpenOrderState",
    "PendingCreationOrderState",
    "PendingCreationChainedOrderState",
    "FillOrderState",
]



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from octobot_trading.personal_data.portfolios cimport portfolio
from octobot_trading.personal_data.portfolios.portfolio cimport (
    Portfolio,
)
from octobot_trading.personal_data.portfolios cimport asset
from octobot_trading.personal_data.portfolios.asset cimport (
    Asset,
)
from octobot_trading.personal_data.portfolios cimport channel
from octobot_trading.personal_data.portfolios.channel cimport (
    BalanceUpdater,
    BalanceProfitabilityUpdater,
    BalanceUpdaterSimulator,
    BalanceProfitabilityUpdaterSimulator,
    BalanceProducer,
    BalanceChannel,
    BalanceProfitabilityProducer,
    BalanceProfitabilityChannel,
)
from octobot_trading.personal_data.portfolios cimport portfolio_factory
from octobot_trading.personal_data.portfolios.portfolio_factory cimport (
    create_portfolio_from_exchange_manager,
)
from octobot_trading.personal_data.portfolios cimport sub_portfolio
from octobot_trading.personal_data.portfolios.sub_portfolio cimport (
    SubPortfolio,
)
from octobot_trading.personal_data.portfolios cimport portfolio_manager
from octobot_trading.personal_data.portfolios.portfolio_manager cimport (
    PortfolioManager,
)
from octobot_trading.personal_data.portfolios cimport types
from octobot_trading.personal_data.portfolios.types cimport (
    FuturePortfolio,
    MarginPortfolio,
    SpotPortfolio,
)
from octobot_trading.personal_data.portfolios.assets cimport (
    FutureAsset,
    MarginAsset,
    SpotAsset,
)
from octobot_trading.personal_data.portfolios cimport portfolio_util
from octobot_trading.personal_data.portfolios.portfolio_util cimport (
    parse_decimal_portfolio,
    filter_empty_values,
    portfolio_to_float,
)
from octobot_trading.personal_data.portfolios cimport history
from octobot_trading.personal_data.portfolios.history cimport (
    create_historical_asset_value_from_dict_like_object,
    HistoricalAssetValue,
    HistoricalPortfolioValueManager,
)

__all__ = [
    "BalanceUpdaterSimulator",
    "BalanceProfitabilityUpdaterSimulator",
    "create_portfolio_from_exchange_manager",
    "BalanceUpdater",
    "BalanceProfitabilityUpdater",
    "Portfolio",
    "Asset",
    "BalanceProducer",
    "BalanceChannel",
    "BalanceProfitabilityProducer",
    "BalanceProfitabilityChannel",
    "SubPortfolio",
    "PortfolioManager",
    "FuturePortfolio",
    "MarginPortfolio",
    "SpotPortfolio",
    "FutureAsset",
    "MarginAsset",
    "SpotAsset",
    "parse_decimal_portfolio",
    "filter_empty_values",
    "portfolio_to_float",
    "create_historical_asset_value_from_dict_like_object",
    "HistoricalAssetValue",
    "HistoricalPortfolioValueManager",
]



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
cimport octobot_trading.exchanges.abstract_exchange as abstract_exchange
cimport octobot_trading.exchanges.config.exchange_config_data as exchange_config_data
cimport octobot_trading.exchanges.abstract_websocket_exchange as abstract_websocket
# cimport octobot_trading.exchange_data.exchange_symbols_data as exchange_symbols_data
# cimport octobot_trading.exchange_data.exchange_symbol_data as exchange_symbol_data
cimport octobot_trading.util as util

cdef class ExchangeManager(util.Initializable):
    cdef public str id
    cdef public str bot_id
    cdef public str exchange_class_string
    cdef public str exchange_name

    cdef public dict config

    cdef public object tentacles_setup_config
    cdef public object logger
    cdef public object backtesting
    cdef public object trader

    cdef public list client_time_frames
    cdef public list client_symbols
    cdef public list trading_modes

    cdef public bint rest_only
    cdef public bint ignore_config
    cdef public bint is_ready
    cdef public bint is_simulated
    cdef public bint is_backtesting
    cdef public bint is_trader_simulated
    cdef public bint is_loading_markets
    cdef public bint is_spot_only
    cdef public bint is_margin
    cdef public bint is_future
    cdef public bint is_sandboxed
    cdef public bint is_trading
    cdef public bint has_websocket
    cdef public bint exchange_only
    cdef public bint without_auth
    cdef public bint check_credentials
    cdef public bint enable_storage

    cdef public object storage_manager
    cdef public abstract_exchange.AbstractExchange exchange
    cdef public abstract_websocket.AbstractWebsocketExchange exchange_web_socket
    cdef public exchange_config_data.ExchangeConfig exchange_config
    cdef public object exchange_backend
    cdef public bint is_valid_account
    cdef public object init_error
    cdef public object exchange_symbols_data
    cdef public object exchange_personal_data

    cdef public dict debug_info

    # private
    cdef object _load_config_symbols_and_time_frames(self)

    # public
    cpdef bint enabled(self)
    cpdef object load_constants(self)
    cpdef str get_exchange_symbol(self, str symbol)
    cpdef tuple get_exchange_quote_and_base(self, str symbol)
    cpdef object get_rest_pairs_refresh_threshold(self)
    cpdef bint need_user_stream(self)
    cpdef void reset_exchange_symbols_data(self)
    cpdef void reset_exchange_personal_data(self)
    cpdef bint check_config(self, str exchange_name)
    cpdef bint symbol_exists(self, str symbol)
    cpdef bint time_frame_exists(self, object time_frame)
    cpdef str get_exchange_name(self)
    cpdef int get_currently_handled_pair_with_time_frame(self)
    cpdef bint get_is_overloaded(self)
    cpdef tuple get_exchange_credentials(self, str exchange_name)
    cpdef str get_exchange_sub_account_id(self, str exchange_name)
    cpdef bint should_decrypt_token(self)
    cpdef object get_symbol_data(self, str symbol)
    cpdef bint is_storage_enabled(self)
    cpdef object update_debug_info(self)
